import logging
import time

import pslab.protocol as CP

logger = logging.getLogger(__name__)


class SPI():
    """
    Methods to interact with the SPI port. An instance of Packet_Handler must be passed to the init function

    """

    def __init__(self, H):
        self.H = H

    def set_parameters(self, primary_prescaler=0, secondary_prescaler=2, CKE=1, CKP=0, SMP=1):
        """
        sets SPI parameters.

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        primary_pres        Primary Prescaler(0,1,2,3) for 64MHz clock->(64:1,16:1,4:1,1:1)
        secondary_pres      Secondary prescaler(0,1,..7)->(8:1,7:1,..1:1)
        CKE                 CKE 0 or 1.
        CKP                 CKP 0 or 1.
        ================    ============================================================================================

        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.SET_SPI_PARAMETERS)
        # 0Bhgfedcba - > <g>: modebit CKP,<f>: modebit CKE, <ed>:primary pre,<cba>:secondary pre
        self.H.send_byte(secondary_prescaler | (primary_prescaler << 3) | (CKE << 5) | (CKP << 6) | (SMP << 7))
        self.H.get_ack()

    def start(self, channel):
        """
        selects SPI channel to enable.
        Basically lowers the relevant chip select pin .

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        channel             1-7 ->[PGA1 connected to CH1,PGA2,PGA3,PGA4,PGA5,external chip select 1,external chip select 2]
                            8 -> sine1
                            9 -> sine2
        ================    ============================================================================================

        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.START_SPI)
        self.H.send_byte(channel)  # value byte
        # self.H.get_ack()

    def set_cs(self, channel, state):
        """
        Enable or disable a chip select

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        channel             'CS1','CS2'
        state               1 for HIGH, 0 for LOW
        ================    ============================================================================================

        """
        channel = channel.upper()
        if channel in ['CS1', 'CS2']:
            csnum = ['CS1', 'CS2'].index(channel) + 9  # chip select number 9=CSOUT1,10=CSOUT2
            self.H.send_byte(CP.SPI_HEADER)
            if state:
                self.H.send_byte(CP.STOP_SPI)
            else:
                self.H.send_byte(CP.START_SPI)
            self.H.send_byte(csnum)
        else:
            print('Channel does not exist')

    def stop(self, channel):
        """
        selects SPI channel to disable.
        Sets the relevant chip select pin to HIGH.

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        channel             1-7 ->[PGA1 connected to CH1,PGA2,PGA3,PGA4,PGA5,external chip select 1,external chip select 2]
        ================    ============================================================================================


        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.STOP_SPI)
        self.H.send_byte(channel)  # value byte
        # self.H.get_ack()

    def send8(self, value):
        """
        SENDS 8-bit data over SPI

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        value               value to transmit
        ================    ============================================================================================

        :return: value returned by slave device
        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.SEND_SPI8)
        self.H.send_byte(value)  # value byte
        v = self.H.get_byte()
        self.H.get_ack()
        return v

    def send16(self, value):
        """
        SENDS 16-bit data over SPI

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        value               value to transmit
        ================    ============================================================================================

        :return: value returned by slave device
        :rtype: int
        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.SEND_SPI16)
        self.H.send_int(value)  # value byte
        v = self.H.get_int()
        self.H.get_ack()
        return v

    def send8_burst(self, value):
        """
        SENDS 8-bit data over SPI
        No acknowledge/return value

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ================    ============================================================================================
        **Arguments**
        ================    ============================================================================================
        value               value to transmit
        ================    ============================================================================================

        :return: Nothing
        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.SEND_SPI8_BURST)
        self.H.send_byte(value)  # value byte

    def send16_burst(self, value):
        """
        SENDS 16-bit data over SPI
        no acknowledge/return value

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ==============  ============================================================================================
        **Arguments**
        ==============  ============================================================================================
        value           value to transmit
        ==============  ============================================================================================

        :return: nothing
        """
        self.H.send_byte(CP.SPI_HEADER)
        self.H.send_byte(CP.SEND_SPI16_BURST)
        self.H.send_int(value)  # value byte

    def xfer(self, chan, data):
        self.start(chan)
        reply = []
        for a in data:
            reply.append(self.send8(a))
        self.stop(chan)
        return reply


class NRF24L01():
    # Commands
    R_REG = 0x00
    W_REG = 0x20
    RX_PAYLOAD = 0x61
    TX_PAYLOAD = 0xA0
    ACK_PAYLOAD = 0xA8
    FLUSH_TX = 0xE1
    FLUSH_RX = 0xE2
    ACTIVATE = 0x50
    R_STATUS = 0xFF

    # Registers
    NRF_CONFIG = 0x00
    EN_AA = 0x01
    EN_RXADDR = 0x02
    SETUP_AW = 0x03
    SETUP_RETR = 0x04
    RF_CH = 0x05
    RF_SETUP = 0x06
    NRF_STATUS = 0x07
    OBSERVE_TX = 0x08
    CD = 0x09
    RX_ADDR_P0 = 0x0A
    RX_ADDR_P1 = 0x0B
    RX_ADDR_P2 = 0x0C
    RX_ADDR_P3 = 0x0D
    RX_ADDR_P4 = 0x0E
    RX_ADDR_P5 = 0x0F
    TX_ADDR = 0x10
    RX_PW_P0 = 0x11
    RX_PW_P1 = 0x12
    RX_PW_P2 = 0x13
    RX_PW_P3 = 0x14
    RX_PW_P4 = 0x15
    RX_PW_P5 = 0x16
    R_RX_PL_WID = 0x60
    FIFO_STATUS = 0x17
    DYNPD = 0x1C
    FEATURE = 0x1D
    PAYLOAD_SIZE = 0
    ACK_PAYLOAD_SIZE = 0
    READ_PAYLOAD_SIZE = 0

    ADC_COMMANDS = 1
    READ_ADC = 0 << 4

    I2C_COMMANDS = 2
    I2C_TRANSACTION = 0 << 4
    I2C_WRITE = 1 << 4
    I2C_SCAN = 2 << 4
    PULL_SCL_LOW = 3 << 4
    I2C_CONFIG = 4 << 4
    I2C_READ = 5 << 4

    NRF_COMMANDS = 3
    NRF_READ_REGISTER = 0
    NRF_WRITE_REGISTER = 1 << 4

    CURRENT_ADDRESS = 0xAAAA01
    nodelist = {}
    nodepos = 0
    NODELIST_MAXLENGTH = 15
    connected = False

    def __init__(self, device):
        self.H = device
        self.ready = False
        self.sigs = {self.CURRENT_ADDRESS: 1}
        if self.H.connected:
            self.connected = self.init()

    def init(self):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_SETUP)
        self.H.get_ack()
        time.sleep(0.015)  # 15 mS settling time
        stat = self.get_status()
        if stat & 0x80:
            logger.info("Radio transceiver not installed/not found")
            return False
        else:
            self.ready = True
        self.selectAddress(self.CURRENT_ADDRESS)
        # self.write_register(self.RF_SETUP,0x06)
        self.rxmode()
        time.sleep(0.1)
        self.flush()
        return True

    def rxmode(self):
        '''
        Puts the radio into listening mode.
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_RXMODE)
        self.H.get_ack()

    def txmode(self):
        '''
        Puts the radio into transmit mode.
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_TXMODE)
        self.H.get_ack()

    def triggerAll(self, val):
        self.txmode()
        self.selectAddress(0x111111)
        self.write_register(self.EN_AA, 0x00)
        self.write_payload([val], True)
        self.write_register(self.EN_AA, 0x01)

    def power_down(self):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_POWER_DOWN)
        self.H.get_ack()

    def rxchar(self):
        '''
        Receives a 1 Byte payload
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_RXCHAR)
        value = self.H.get_byte()
        self.H.get_ack()
        return value

    def txchar(self, char):
        '''
        Transmits a single character
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_TXCHAR)
        self.H.send_byte(char)
        return self.H.get_ack() >> 4

    def hasData(self):
        '''
        Check if the RX FIFO contains data
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_HASDATA)
        value = self.H.get_byte()
        self.H.get_ack()
        return value

    def flush(self):
        '''
        Flushes the TX and RX FIFOs
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_FLUSH)
        self.H.get_ack()

    def write_register(self, address, value):
        '''
        write a  byte to any of the configuration registers on the Radio.
        address byte can either be located in the NRF24L01+ manual, or chosen
        from some of the constants defined in this module.
        '''
        # print ('writing',address,value)
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITEREG)
        self.H.send_byte(address)
        self.H.send_byte(value)
        self.H.get_ack()

    def read_register(self, address):
        '''
        Read the value of any of the configuration registers on the radio module.

        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_READREG)
        self.H.send_byte(address)
        val = self.H.get_byte()
        self.H.get_ack()
        return val

    def get_status(self):
        '''
        Returns a byte representing the STATUS register on the radio.
        Refer to NRF24L01+ documentation for further details
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_GETSTATUS)
        val = self.H.get_byte()
        self.H.get_ack()
        return val

    def write_command(self, cmd):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITECOMMAND)
        self.H.send_byte(cmd)
        self.H.get_ack()

    def write_address(self, register, address):
        '''
        register can be TX_ADDR, RX_ADDR_P0 -> RX_ADDR_P5
        3 byte address.  eg 0xFFABXX . XX cannot be FF
        if RX_ADDR_P1 needs to be used along with any of the pipes
        from P2 to P5, then RX_ADDR_P1 must be updated last.
        Addresses from P1-P5 must share the first two bytes.
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITEADDRESS)
        self.H.send_byte(register)
        self.H.send_byte(address & 0xFF)
        self.H.send_byte((address >> 8) & 0xFF)
        self.H.send_byte((address >> 16) & 0xFF)
        self.H.get_ack()

    def selectAddress(self, address):
        '''
        Sets RX_ADDR_P0 and TX_ADDR to the specified address.

        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITEADDRESSES)
        self.H.send_byte(address & 0xFF)
        self.H.send_byte((address >> 8) & 0xFF)
        self.H.send_byte((address >> 16) & 0xFF)
        self.H.get_ack()
        self.CURRENT_ADDRESS = address
        if address not in self.sigs:
            self.sigs[address] = 1

    def read_payload(self, numbytes):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_READPAYLOAD)
        self.H.send_byte(numbytes)
        data = self.H.fd.read(numbytes)
        self.H.get_ack()
        return [ord(a) for a in data]

    def write_payload(self, data, verbose=False, **args):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITEPAYLOAD)
        numbytes = len(
            data) | 0x80  # 0x80 implies transmit immediately. Otherwise it will simply load the TX FIFO ( used by ACK_payload)
        if (args.get('rxmode', False)): numbytes |= 0x40
        self.H.send_byte(numbytes)
        self.H.send_byte(self.TX_PAYLOAD)
        for a in data:
            self.H.send_byte(a)
        val = self.H.get_ack() >> 4
        if (verbose):
            if val & 0x2:
                print(' NRF radio not found. Connect one to the add-on port')
            elif val & 0x1:
                print(' Node probably dead/out of range. It failed to acknowledge')
            return
        return val

    def I2C_scan(self):
        '''
        Scans the I2C bus and returns a list of live addresses
        '''
        x = self.transaction([self.I2C_COMMANDS | self.I2C_SCAN | 0x80], timeout=500)
        if not x: return []
        if not sum(x): return []
        addrs = []
        for a in range(16):
            if (x[a] ^ 255):
                for b in range(8):
                    if x[a] & (0x80 >> b) == 0:
                        addr = 8 * a + b
                        addrs.append(addr)
        return addrs

    def GuessingScan(self):
        '''
        Scans the I2C bus and also prints the possible devices associated with each found address
        '''
        from PSL import sensorlist
        print('Scanning addresses 0-127...')
        x = self.transaction([self.I2C_COMMANDS | self.I2C_SCAN | 0x80], timeout=500)
        if not x: return []
        if not sum(x): return []
        addrs = []
        print('Address', '\t', 'Possible Devices')

        for a in range(16):
            if (x[a] ^ 255):
                for b in range(8):
                    if x[a] & (0x80 >> b) == 0:
                        addr = 8 * a + b
                        addrs.append(addr)
                        print(hex(addr), '\t\t', sensorlist.sensors.get(addr, 'None'))

        return addrs

    def transaction(self, data, **args):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_TRANSACTION)
        self.H.send_byte(len(data))  # total Data bytes coming through
        if 'listen' not in args: args['listen'] = True
        if args.get('listen', False): data[0] |= 0x80  # You need this if hardware must wait for a reply
        timeout = args.get('timeout', 200)
        verbose = args.get('verbose', False)
        self.H.send_int(timeout)  # timeout.
        for a in data:
            self.H.send_byte(a)

        # print ('dt send',time.time()-st,timeout,data[0]&0x80,data)
        numbytes = self.H.get_byte()
        # print ('byte 1 in',time.time()-st)
        if numbytes:
            data = self.H.fd.read(numbytes)
        else:
            data = []
        val = self.H.get_ack() >> 4
        if (verbose):
            if val & 0x1: print(time.time(), '%s Err. Node not found' % (hex(self.CURRENT_ADDRESS)))
            if val & 0x2: print(time.time(),
                                '%s Err. NRF on-board transmitter not found' % (hex(self.CURRENT_ADDRESS)))
            if val & 0x4 and args['listen']: print(time.time(),
                                                   '%s Err. Node received command but did not reply' % (
                                                       hex(self.CURRENT_ADDRESS)))
        if val & 0x7:  # Something didn't go right.
            self.flush()
            self.sigs[self.CURRENT_ADDRESS] = self.sigs[self.CURRENT_ADDRESS] * 50 / 51.
            return False

        self.sigs[self.CURRENT_ADDRESS] = (self.sigs[self.CURRENT_ADDRESS] * 50 + 1) / 51.
        return [ord(a) for a in data]

    def transactionWithRetries(self, data, **args):
        retries = args.get('retries', 5)
        reply = False
        while retries > 0:
            reply = self.transaction(data, verbose=(retries == 1), **args)
            if reply:
                break
            retries -= 1
        return reply

    def write_ack_payload(self, data, pipe):
        if (len(data) != self.ACK_PAYLOAD_SIZE):
            self.ACK_PAYLOAD_SIZE = len(data)
            if self.ACK_PAYLOAD_SIZE > 15:
                print('too large. truncating.')
                self.ACK_PAYLOAD_SIZE = 15
                data = data[:15]
            else:
                print('ack payload size:', self.ACK_PAYLOAD_SIZE)

        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_WRITEPAYLOAD)
        self.H.send_byte(len(data))
        self.H.send_byte(self.ACK_PAYLOAD | pipe)
        for a in data:
            self.H.send_byte(a)
        return self.H.get_ack() >> 4

    def start_token_manager(self):
        '''
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_START_TOKEN_MANAGER)
        self.H.get_ack()

    def stop_token_manager(self):
        '''
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_STOP_TOKEN_MANAGER)
        self.H.get_ack()

    def total_tokens(self):
        '''
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_TOTAL_TOKENS)
        x = self.H.get_byte()
        self.H.get_ack()
        return x

    def fetch_report(self, num):
        '''
        '''
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_REPORTS)
        self.H.send_byte(num)
        data = [self.H.get_byte() for a in range(20)]
        self.H.get_ack()
        return data

    @staticmethod
    def __decode_I2C_list__(data):
        lst = []
        if sum(data) == 0:
            return lst
        for i, d in enumerate(data):
            if (d ^ 255):
                for b in range(8):
                    if d & (0x80 >> b) == 0:
                        addr = 8 * i + b
                        lst.append(addr)
        return lst

    def get_nodelist(self):
        '''
        Refer to the variable 'nodelist' if you simply want a list of nodes that either registered while your code was
        running , or were loaded from the firmware buffer(max 15 entries)

        If you plan to use more than 15 nodes, and wish to register their addresses without having to feed them manually,
        then this function must be called each time before the buffer resets.

        The dictionary object returned by this function [addresses paired with arrays containing their registered sensors]
        is filtered by checking with each node if they are alive.

        '''

        total = self.total_tokens()
        if self.nodepos != total:
            for nm in range(self.NODELIST_MAXLENGTH):
                dat = self.fetch_report(nm)
                txrx = (dat[0]) | (dat[1] << 8) | (dat[2] << 16)
                if not txrx: continue
                self.nodelist[txrx] = self.__decode_I2C_list__(dat[3:19])
                self.nodepos = total
                # else:
                #	self.__delete_registered_node__(nm)

        filtered_lst = {}
        for a in self.nodelist:
            if self.isAlive(a): filtered_lst[a] = self.nodelist[a]

        return filtered_lst

    def __delete_registered_node__(self, num):
        self.H.send_byte(CP.NRFL01)
        self.H.send_byte(CP.NRF_DELETE_REPORT_ROW)
        self.H.send_byte(num)
        self.H.get_ack()

    def __delete_all_registered_nodes__(self):
        while self.total_tokens():
            print('-')
            self.__delete_registered_node__(0)

    def isAlive(self, addr):
        self.selectAddress(addr)
        return self.transaction([self.NRF_COMMANDS | self.NRF_READ_REGISTER] + [self.R_STATUS], timeout=100,
                                verbose=False)

    def init_shockburst_transmitter(self, **args):
        '''
        Puts the radio into transmit mode.
        Dynamic Payload with auto acknowledge is enabled.
        upto 5 retransmits with 1ms delay between each in case a node doesn't respond in time
        Receivers must acknowledge payloads
        '''
        self.PAYLOAD_SIZE = args.get('PAYLOAD_SIZE', self.PAYLOAD_SIZE)
        myaddr = args.get('myaddr', 0xAAAA01)
        sendaddr = args.get('sendaddr', 0xAAAA01)

        self.init()
        # shockburst
        self.write_address(self.RX_ADDR_P0, myaddr)  # transmitter's address
        self.write_address(self.TX_ADDR, sendaddr)  # send to node with this address
        self.write_register(self.RX_PW_P0, self.PAYLOAD_SIZE)
        self.rxmode()
        time.sleep(0.1)
        self.flush()

    def init_shockburst_receiver(self, **args):
        '''
        Puts the radio into receive mode.
        Dynamic Payload with auto acknowledge is enabled.
        '''
        self.PAYLOAD_SIZE = args.get('PAYLOAD_SIZE', self.PAYLOAD_SIZE)
        if 'myaddr0' not in args:
            args['myaddr0'] = 0xA523B5
        # if 'sendaddr' non in args:
        #	args['sendaddr']=0xA523B5
        print(args)
        self.init()
        self.write_register(self.RF_SETUP, 0x26)  # 2MBPS speed

        # self.write_address(self.TX_ADDR,sendaddr)     #send to node with this address
        # self.write_address(self.RX_ADDR_P0,myaddr)	#will receive the ACK Payload from that node
        enabled_pipes = 0  # pipes to be enabled
        for a in range(0, 6):
            x = args.get('myaddr' + str(a), None)
            if x:
                print(hex(x), hex(self.RX_ADDR_P0 + a))
                enabled_pipes |= (1 << a)
                self.write_address(self.RX_ADDR_P0 + a, x)
        P15_base_address = args.get('myaddr1', None)
        if P15_base_address: self.write_address(self.RX_ADDR_P1, P15_base_address)

        self.write_register(self.EN_RXADDR, enabled_pipes)  # enable pipes
        self.write_register(self.EN_AA, enabled_pipes)  # enable auto Acknowledge on all pipes
        self.write_register(self.DYNPD, enabled_pipes)  # enable dynamic payload on Data pipes
        self.write_register(self.FEATURE, 0x06)  # enable dynamic payload length
        # self.write_register(self.RX_PW_P0,self.PAYLOAD_SIZE)

        self.rxmode()
        time.sleep(0.1)
        self.flush()


class RadioLink():
    ADC_COMMANDS = 1
    READ_ADC = 0 << 4

    I2C_COMMANDS = 2
    I2C_TRANSACTION = 0 << 4
    I2C_WRITE = 1 << 4
    SCAN_I2C = 2 << 4
    PULL_SCL_LOW = 3 << 4
    I2C_CONFIG = 4 << 4
    I2C_READ = 5 << 4

    NRF_COMMANDS = 3
    NRF_READ_REGISTER = 0 << 4
    NRF_WRITE_REGISTER = 1 << 4

    MISC_COMMANDS = 4
    WS2812B_CMD = 0 << 4

    def __init__(self, NRF, **args):
        self.NRF = NRF
        if 'address' in args:
            self.ADDRESS = args.get('address', False)
        else:
            print('Address not specified. Add "address=0x....." argument while instantiating')
            self.ADDRESS = 0x010101

    def __selectMe__(self):
        if self.NRF.CURRENT_ADDRESS != self.ADDRESS:
            self.NRF.selectAddress(self.ADDRESS)

    def I2C_scan(self):
        self.__selectMe__()
        from PSL import sensorlist
        print('Scanning addresses 0-127...')
        x = self.NRF.transaction([self.I2C_COMMANDS | self.SCAN_I2C | 0x80], timeout=500)
        if not x: return []
        if not sum(x): return []
        addrs = []
        print('Address', '\t', 'Possible Devices')

        for a in range(16):
            if (x[a] ^ 255):
                for b in range(8):
                    if x[a] & (0x80 >> b) == 0:
                        addr = 8 * a + b
                        addrs.append(addr)
                        print(hex(addr), '\t\t', sensorlist.sensors.get(addr, 'None'))

        return addrs

    @staticmethod
    def __decode_I2C_list__(data):
        lst = []
        if sum(data) == 0:
            return lst
        for i, d in enumerate(data):
            if (d ^ 255):
                for b in range(8):
                    if d & (0x80 >> b) == 0:
                        addr = 8 * i + b
                        lst.append(addr)
        return lst

    def writeI2C(self, I2C_addr, regaddress, data_bytes):
        self.__selectMe__()
        return self.NRF.transaction([self.I2C_COMMANDS | self.I2C_WRITE] + [I2C_addr] + [regaddress] + data_bytes)

    def readI2C(self, I2C_addr, regaddress, numbytes):
        self.__selectMe__()
        return self.NRF.transaction([self.I2C_COMMANDS | self.I2C_TRANSACTION] + [I2C_addr] + [regaddress] + [numbytes])

    def writeBulk(self, I2C_addr, data_bytes):
        self.__selectMe__()
        return self.NRF.transaction([self.I2C_COMMANDS | self.I2C_WRITE] + [I2C_addr] + data_bytes)

    def readBulk(self, I2C_addr, regaddress, numbytes):
        self.__selectMe__()
        return self.NRF.transactionWithRetries(
            [self.I2C_COMMANDS | self.I2C_TRANSACTION] + [I2C_addr] + [regaddress] + [numbytes])

    def simpleRead(self, I2C_addr, numbytes):
        self.__selectMe__()
        return self.NRF.transactionWithRetries([self.I2C_COMMANDS | self.I2C_READ] + [I2C_addr] + [numbytes])

    def readADC(self, channel):
        self.__selectMe__()
        return self.NRF.transaction([self.ADC_COMMANDS | self.READ_ADC] + [channel])

    def pullSCLLow(self, t_ms):
        self.__selectMe__()
        dat = self.NRF.transaction([self.I2C_COMMANDS | self.PULL_SCL_LOW] + [t_ms])
        if dat:
            return self.__decode_I2C_list__(dat)
        else:
            return []

    def configI2C(self, freq):
        self.__selectMe__()
        brgval = int(32e6 / freq / 4 - 1)
        print(brgval)
        return self.NRF.transaction([self.I2C_COMMANDS | self.I2C_CONFIG] + [brgval], listen=False)

    def write_register(self, reg, val):
        self.__selectMe__()
        # print ('writing to ',reg,val)
        return self.NRF.transaction([self.NRF_COMMANDS | self.NRF_WRITE_REGISTER] + [reg, val], listen=False)

    def WS2812B(self, cols):
        """
        set shade of WS2182 LED on CS1/RC0

        .. tabularcolumns:: |p{3cm}|p{11cm}|

        ==============  ============================================================================================
        **Arguments**
        ==============  ============================================================================================
        cols                2Darray [[R,G,B],[R2,G2,B2],[R3,G3,B3]...]
                            brightness of R,G,B ( 0-255  )
        ==============  ============================================================================================

        example::

            >>> WS2812B([[10,0,0],[0,10,10],[10,0,10]])
            #sets red, cyan, magenta to three daisy chained LEDs

        """
        self.__selectMe__()
        colarray = []
        for a in cols:
            colarray.append(int('{:08b}'.format(int(a[1]))[::-1], 2))
            colarray.append(int('{:08b}'.format(int(a[0]))[::-1], 2))
            colarray.append(int('{:08b}'.format(int(a[2]))[::-1], 2))

        res = self.NRF.transaction([self.MISC_COMMANDS | self.WS2812B_CMD] + colarray, listen=False)
        return res

    def read_register(self, reg):
        self.__selectMe__()
        x = self.NRF.transaction([self.NRF_COMMANDS | self.NRF_READ_REGISTER] + [reg])
        if x:
            return x[0]
        else:
            return False



# -*- coding: utf-8 -*-
# MF522 - Software stack to access the MF522 RFID reader via  FOSSASIA PSLab
#
from __future__ import print_function


def connect(I, cs):
    return MF522(I, cs)


class MF522:
    # Constants from https://github.com/miguelbalboa/rfid ( Open source License : UNLICENSE)
    CommandReg = 0x01 << 1  # starts and stops command execution
    ComIEnReg = 0x02 << 1  # enable and disable interrupt request control bits
    DivIEnReg = 0x03 << 1  # enable and disable interrupt request control bits
    ComIrqReg = 0x04 << 1  # interrupt request bits
    DivIrqReg = 0x05 << 1  # interrupt request bits
    ErrorReg = 0x06 << 1  # error bits showing the error status of the last command executed
    Status1Reg = 0x07 << 1  # communication status bits
    Status2Reg = 0x08 << 1  # receiver and transmitter status bits
    FIFODataReg = 0x09 << 1  # input and output of 64 byte FIFO buffer
    FIFOLevelReg = 0x0A << 1  # number of bytes stored in the FIFO buffer
    WaterLevelReg = 0x0B << 1  # level for FIFO underflow and overflow warning
    ControlReg = 0x0C << 1  # miscellaneous control registers
    BitFramingReg = 0x0D << 1  # adjustments for bit-oriented frames
    CollReg = 0x0E << 1  # bit position of the first bit-collision detected on the RF sciencelab

    ModeReg = 0x11 << 1  # defines general modes for transmitting and receiving
    TxModeReg = 0x12 << 1  # defines transmission data rate and framing
    RxModeReg = 0x13 << 1  # defines reception data rate and framing
    TxControlReg = 0x14 << 1  # controls the logical behavior of the antenna driver pins TX1 and TX2
    TxASKReg = 0x15 << 1  # controls the setting of the transmission modulation
    TxSelReg = 0x16 << 1  # selects the internal sources for the antenna driver
    RxSelReg = 0x17 << 1  # selects internal receiver settings
    RxThresholdReg = 0x18 << 1  # selects thresholds for the bit decoder
    DemodReg = 0x19 << 1  # defines demodulator settings
    MfTxReg = 0x1C << 1  # controls some MIFARE communication transmit parameters
    MfRxReg = 0x1D << 1  # controls some MIFARE communication receive parameters
    SerialSpeedReg = 0x1F << 1  # selects the speed of the serial UART sciencelab

    CRCResultRegH = 0x21 << 1  # shows the MSB and LSB values of the CRC calculation
    CRCResultRegL = 0x22 << 1
    ModWidthReg = 0x24 << 1  # controls the ModWidth setting?
    RFCfgReg = 0x26 << 1  # configures the receiver gain
    GsNReg = 0x27 << 1  # selects the conductance of the antenna driver pins TX1 and TX2 for modulation
    CWGsPReg = 0x28 << 1  # defines the conductance of the p-driver output during periods of no modulation
    ModGsPReg = 0x29 << 1  # defines the conductance of the p-driver output during periods of modulation
    TModeReg = 0x2A << 1  # defines settings for the internal timer
    TPrescalerReg = 0x2B << 1  # the lower 8 bits of the TPrescaler value. The 4 high bits are in TModeReg.
    TReloadRegH = 0x2C << 1  # defines the 16-bit timer reload value
    TReloadRegL = 0x2D << 1
    TCounterValueRegH = 0x2E << 1  # shows the 16-bit timer value
    TCounterValueRegL = 0x2F << 1

    TestSel1Reg = 0x31 << 1  # general test signal configuration
    TestSel2Reg = 0x32 << 1  # general test signal configuration
    TestPinEnReg = 0x33 << 1  # enables pin output driver on pins D1 to D7
    TestPinValueReg = 0x34 << 1  # defines the values for D1 to D7 when it is used as an I/O bus
    TestBusReg = 0x35 << 1  # shows the status of the internal test bus
    AutoTestReg = 0x36 << 1  # controls the digital self test
    VersionReg = 0x37 << 1  # shows the software version
    AnalogTestReg = 0x38 << 1  # controls the pins AUX1 and AUX2
    TestDAC1Reg = 0x39 << 1  # defines the test value for TestDAC1
    TestDAC2Reg = 0x3A << 1  # defines the test value for TestDAC2
    TestADCReg = 0x3B << 1  # shows the value of ADC I and Q channels

    # MFRC522 commands. Described in chapter 10 of the datasheet.
    PCD_Idle = 0x00  # no action, cancels current command execution
    PCD_Mem = 0x01  # stores 25 bytes into the internal buffer
    PCD_GenerateRandomID = 0x02  # generates a 10-byte random ID number
    PCD_CalcCRC = 0x03  # activates the CRC coprocessor or performs a self test
    PCD_Transmit = 0x04  # transmits data from the FIFO buffer
    PCD_NoCmdChange = 0x07
    PCD_Receive = 0x08  # activates the receiver circuits
    PCD_Transceive = 0x0C  # transmits data from FIFO buffer to antenna and automatically activates the receiver after transmission
    PCD_MFAuthent = 0x0E  # performs the MIFARE standard authentication as a reader
    PCD_SoftReset = 0x0F  # resets the MFRC522

    RxGain_18dB = 0x00 << 4  # 000b - 18 dB, minimum
    RxGain_23dB = 0x01 << 4  # 001b - 23 dB
    RxGain_18dB_2 = 0x02 << 4  # 010b - 18 dB, it seems 010b is a duplicate for 000b
    RxGain_23dB_2 = 0x03 << 4  # 011b - 23 dB, it seems 011b is a duplicate for 001b
    RxGain_33dB = 0x04 << 4  # 100b - 33 dB, average, and typical default
    RxGain_38dB = 0x05 << 4  # 101b - 38 dB
    RxGain_43dB = 0x06 << 4  # 110b - 43 dB
    RxGain_48dB = 0x07 << 4  # 111b - 48 dB, maximum
    RxGain_min = 0x00 << 4  # 000b - 18 dB, minimum, convenience for RxGain_18dB
    RxGain_avg = 0x04 << 4  # 100b - 33 dB, average, convenience for RxGain_33dB
    RxGain_max = 0x07 << 4  # 111b - 48 dB, maximum, convenience for RxGain_48dB

    # The commands used by the PCD to manage communication with several PICCs (ISO 14443-3, Type A, section 6.4)
    PICC_CMD_REQA = 0x26  # REQuest command, Type A. Invites PICCs in state IDLE to go to READY and prepare for anticollision or selection
    PICC_CMD_WUPA = 0x52  # Wake-UP command, prepare for anticollision or selection. 7 bit frame.
    PICC_CMD_CT = 0x88  # Cascade Tag. Not really a command, but used during anti collision.
    PICC_CMD_SEL_CL1 = 0x93  # Anti collision/Select, Cascade Level 1
    PICC_CMD_SEL_CL2 = 0x95  # Anti collision/Select, Cascade Level 2
    PICC_CMD_SEL_CL3 = 0x97  # Anti collision/Select, Cascade Level 3
    PICC_CMD_HLTA = 0x50  # HaLT command, Type A. Instructs an ACTIVE PICC to go to state HALT.
    # The commands used for MIFARE Classic (from http://www.mouser.com/ds/2/302/MF1S503x-89574.pdf, Section 9)
    # Use PCD_MFAuthent to authenticate access to a sector, then use these commands to read/write/modify the blocks on the sector.
    # The read/write commands can also be used for MIFARE Ultralight.
    PICC_CMD_MF_AUTH_KEY_A = 0x60  # Perform authentication with Key A
    PICC_CMD_MF_AUTH_KEY_B = 0x61  # Perform authentication with Key B
    PICC_CMD_MF_READ = 0x30  # Reads one 16 byte block from the authenticated sector of the PICC. Also used for MIFARE Ultralight.
    PICC_CMD_MF_WRITE = 0xA0  # Writes one 16 byte block to the authenticated sector of the PICC. Called "COMPATIBILITY WRITE" for MIFARE Ultralight.
    PICC_CMD_MF_DECREMENT = 0xC0  # Decrements the contents of a block and stores the result in the internal data register.
    PICC_CMD_MF_INCREMENT = 0xC1  # Increments the contents of a block and stores the result in the internal data register.
    PICC_CMD_MF_RESTORE = 0xC2  # Reads the contents of a block into the internal data register.
    PICC_CMD_MF_TRANSFER = 0xB0  # Writes the contents of the internal data register to a block.

    NRSTPD = 22
    MAX_LEN = 16
    MI_OK = 0
    MI_NOTAGERR = 1
    MI_ERR = 2

    PCD_CALCCRC = 0x03

    PICC_REQIDL = 0x26
    PICC_REQALL = 0x52
    PICC_ANTICOLL = 0x93
    PICC_SElECTTAG = 0x93
    PICC_AUTHENT1A = 0x60
    PICC_AUTHENT1B = 0x61
    PICC_READ = 0x30
    PICC_WRITE = 0xA0
    PICC_DECREMENT = 0xC0
    PICC_INCREMENT = 0xC1
    PICC_RESTORE = 0xC2
    PICC_TRANSFER = 0xB0
    PICC_HALT = 0x50

    # The commands used for MIFARE Ultralight (from http://www.nxp.com/documents/data_sheet/MF0ICU1.pdf, Section 8.6)
    # The PICC_CMD_MF_READ and PICC_CMD_MF_WRITE can also be used for MIFARE Ultralight.
    PICC_CMD_UL_WRITE = 0xA2  # Writes one 4 byte page to the PICC.

    MF_ACK = 0xA  # The MIFARE Classic uses a 4 bit ACK/NAK. Any other value than 0xA is NAK.
    MF_KEY_SIZE = 6  # A Mifare Crypto1 key is 6 bytes.

    def __init__(self, I, cs='CS1'):
        self.cs = cs
        self.I = I
        self.I.SPI.set_parameters(2, 1, 1, 0)
        if not self.reset():
            self.connected = False
            return None
        self.write(self.TModeReg, 0x80)
        self.write(self.TPrescalerReg, 0xA9)
        self.write(self.TReloadRegH, 0x03)
        self.write(self.TReloadRegL, 0xE8)

        self.write(self.TxASKReg, 0x40)
        self.write(self.ModeReg, 0x3D)

        # Enable the antenna
        self.enableAntenna()
        self.connected = True

    def enableAntenna(self):
        val = self.read(self.TxControlReg);
        if ((val & 0x03) != 0x03):
            self.write(self.TxControlReg, val | 0x03);

    def reset(self):
        self.write(self.CommandReg, self.PCD_SoftReset)
        s = time.time()
        while (self.read(self.CommandReg) & (1 << 4)):
            print('wait')
            time.sleep(0.1)
            if time.time() - s > .5: return False
        return True

    def write(self, register, val):
        self.I.SPI.set_cs(self.cs, 0)
        ret = self.I.SPI.send16(((register & 0x7F) << 8) | val)
        self.I.SPI.set_cs(self.cs, 1)
        return ret & 0xFF

    def read(self, register):
        self.I.SPI.set_cs(self.cs, 0)
        ret = self.I.SPI.send16((register | 0x80) << 8)
        self.I.SPI.set_cs(self.cs, 1)
        return ret & 0xFF

    def readMany(self, register, total):
        self.I.SPI.set_cs(self.cs, 0)
        self.I.SPI.send8(register)
        vals = []
        for a in range(total - 1):
            vals.append(I.SPI.send8(register))
        vals.append(I.SPI.send8(0))
        self.I.SPI.set_cs(self.cs, 1)
        return vals

    def getStatus(self):
        return self.read(self.Status1Reg)

    def getVersion(self):
        ver = self.read(self.VersionReg)
        if ver == 0x88:
            print('Cloned version: Fudan Semiconductors')
        elif ver == 0x90:
            print('version 1.0')
        elif ver == 0x91:
            print('version 1.0')
        elif ver == 0x92:
            print('version 2.0')
        else:
            print('Unknown version ', ver)
        return ver

    def SetBitMask(self, reg, mask):
        tmp = self.read(reg)
        self.write(reg, tmp | mask)

    def ClearBitMask(self, reg, mask):
        tmp = self.read(reg);
        self.write(reg, tmp & (~mask))

    def MFRC522_ToCard(self, command, sendData):
        returnedData = []
        backLen = 0
        status = self.MI_ERR
        irqEn = 0x00
        waitIRq = 0x00
        lastBits = None
        n = 0
        i = 0

        if command == self.PCD_MFAuthent:
            irqEn = 0x12
            waitIRq = 0x10
        if command == self.PCD_Transceive:
            irqEn = 0x77
            waitIRq = 0x30

        self.write(self.ComIEnReg, irqEn | 0x80)
        self.ClearBitMask(self.ComIrqReg, 0x80)
        self.SetBitMask(self.FIFOLevelReg, 0x80)

        self.write(self.CommandReg, self.PCD_Idle);

        for a in sendData:
            self.write(self.FIFODataReg, a)
        self.write(self.CommandReg, command)

        if command == self.PCD_Transceive:
            self.SetBitMask(self.BitFramingReg, 0x80)

        i = 2000
        while True:
            n = self.read(self.ComIrqReg)
            i = i - 1
            if ~((i != 0) and ~(n & 0x01) and ~(n & waitIRq)):
                break

        self.ClearBitMask(self.BitFramingReg, 0x80)

        if i != 0:
            if (self.read(self.ErrorReg) & 0x1B) == 0x00:
                status = self.MI_OK

                if n & irqEn & 0x01:
                    status = self.MI_NOTAGERR

                if command == self.PCD_Transceive:
                    n = self.read(self.FIFOLevelReg)
                    lastBits = self.read(self.ControlReg) & 0x07
                    if lastBits != 0:
                        backLen = (n - 1) * 8 + lastBits
                    else:
                        backLen = n * 8

                    if n == 0:
                        n = 1
                    if n > self.MAX_LEN:
                        n = self.MAX_LEN

                    i = 0
                    while i < n:
                        returnedData.append(self.read(self.FIFODataReg))
                        i = i + 1;
            else:
                status = self.MI_ERR
        return (status, returnedData, backLen)

    def MFRC522_Request(self, reqMode):
        status = None
        backBits = None
        TagType = []

        self.write(self.BitFramingReg, 0x07)

        TagType.append(reqMode);
        (status, returnedData, backBits) = self.MFRC522_ToCard(self.PCD_Transceive, TagType)

        if ((status != self.MI_OK) | (backBits != 0x10)):
            status = self.MI_ERR
        return (status, backBits)

    def MFRC522_Anticoll(self):
        returnedData = []
        serNumCheck = 0

        serNum = []

        self.write(self.BitFramingReg, 0x00)

        serNum.append(self.PICC_ANTICOLL)
        serNum.append(0x20)

        (status, returnedData, backBits) = self.MFRC522_ToCard(self.PCD_Transceive, serNum)

        if (status == self.MI_OK):
            i = 0
            if len(returnedData) == 5:
                while i < 4:
                    serNumCheck = serNumCheck ^ returnedData[i]
                    i = i + 1
                if serNumCheck != returnedData[i]:
                    status = self.MI_ERR
            else:
                status = self.MI_ERR

        return (status, returnedData)

    def CalulateCRC(self, pIndata):
        self.ClearBitMask(self.DivIrqReg, 0x04)
        self.SetBitMask(self.FIFOLevelReg, 0x80);
        for a in pIndata:
            self.write(self.FIFODataReg, a)
        self.write(self.CommandReg, self.PCD_CALCCRC)
        for i in range(0xFF):
            n = self.read(self.DivIrqReg)
            if (n & 0x04):
                break
        pOutData = []
        pOutData.append(self.read(self.CRCResultRegL))
        pOutData.append(self.read(self.CRCResultRegH))
        return pOutData

    def MFRC522_SelectTag(self, serNum):
        returnedData = []
        buf = []
        buf.append(self.PICC_SElECTTAG)
        buf.append(0x70)
        i = 0
        while i < 5:
            buf.append(serNum[i])
            i = i + 1
        pOut = self.CalulateCRC(buf)
        buf.append(pOut[0])
        buf.append(pOut[1])
        (status, returnedData, backLen) = self.MFRC522_ToCard(self.PCD_Transceive, buf)

        if (status == self.MI_OK) and (backLen == 0x18):
            return returnedData[0]
        else:
            return 0

    def MFRC522_Auth(self, authMode, BlockAddr, Sectorkey, serNum):
        buff = []
        # First byte should be the authMode (A or B)
        buff.append(authMode)
        # Second byte is the trailerBlock (usually 7)
        buff.append(BlockAddr)
        # Now we need to append the authKey which usually is 6 bytes of 0xFF
        i = 0
        while (i < len(Sectorkey)):
            buff.append(Sectorkey[i])
            i = i + 1
        i = 0

        # Next we append the first 4 bytes of the UID
        while (i < 4):
            buff.append(serNum[i])
            i = i + 1

        # Now we start the authentication itself
        (status, returnedData, backLen) = self.MFRC522_ToCard(self.PCD_MFAuthent, buff)

        # Check if an error occurred
        if not (status == self.MI_OK):
            print("AUTH ERROR!!")
        if not (self.read(self.Status2Reg) & 0x08) != 0:
            print("AUTH ERROR(status2reg & 0x08) != 0")

        # Return the status
        return status

    def MFRC522_StopCrypto1(self):
        self.ClearBitMask(self.Status2Reg, 0x08)
        self.SetBitMask(self.CommandReg, 0x10)

    def MFRC522_Read(self, blockAddr):
        recvData = []
        recvData.append(self.PICC_READ)
        recvData.append(blockAddr)
        pOut = self.CalulateCRC(recvData)
        recvData.append(pOut[0])
        recvData.append(pOut[1])
        (status, returnedData, backLen) = self.MFRC522_ToCard(self.PCD_Transceive, recvData)
        if not (status == self.MI_OK):
            print("Error while reading!")
        i = 0
        return returnedData

    def MFRC522_Write(self, blockAddr, writeData):
        buff = []
        buff.append(self.PICC_WRITE)
        buff.append(blockAddr)
        crc = self.CalulateCRC(buff)
        buff.append(crc[0])
        buff.append(crc[1])
        (status, returnedData, backLen) = self.MFRC522_ToCard(self.PCD_Transceive, buff)
        if not (status == self.MI_OK) or not (backLen == 4) or not ((returnedData[0] & 0x0F) == 0x0A):
            status = self.MI_ERR

        print(str(backLen) + " returnedData &0x0F == 0x0A " + str(returnedData[0] & 0x0F))
        if status == self.MI_OK:
            i = 0
            buf = []
            while i < 16:
                buf.append(writeData[i])
                i = i + 1
            crc = self.CalulateCRC(buf)
            buf.append(crc[0])
            buf.append(crc[1])
            (status, returnedData, backLen) = self.MFRC522_ToCard(self.PCD_Transceive, buf)
            if not (status == self.MI_OK) or not (backLen == 4) or not ((returnedData[0] & 0x0F) == 0x0A):
                print("Error while writing")
            if status == self.MI_OK:
                print("Data written")

    def MFRC522_DumpClassic1K(self, key, uid):
        i = 0
        while i < 64:
            status = self.MFRC522_Auth(self.PICC_AUTHENT1A, i, key, uid)
            # Check if authenticated
            if status == self.MI_OK:
                self.MFRC522_Read(i)
            else:
                print("Authentication error")
            i = i + 1


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    A = MF522(I, 'CS1')
    ret = A.getStatus()
    print(ret, hex(ret), bin(ret))
    A.getVersion()
    import time

    while 1:
        (status, TagType) = A.MFRC522_Request(A.PICC_CMD_REQA)
        if status == A.MI_OK:
            print("Card detected")
            (status, uid) = A.MFRC522_Anticoll()
            if status == A.MI_OK:
                print("Card read UID: " + str(uid[0]) + "," + str(uid[1]) + "," + str(uid[2]) + "," + str(uid[3]))
                key = [0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]
                A.MFRC522_SelectTag(uid)
                status = A.MFRC522_Auth(A.PICC_AUTHENT1A, 8, key, uid)
                if status == A.MI_OK:
                    print(A.MFRC522_Read(8))
                    '''
                    # Variable for the data to write
                    data = []
                    # Fill the data with 0xFF
                    for x in range(0,16):
                        data.append(0xFF)
                    print ("Sector 8 looked like this:")
                    # Read block 8
                    A.MFRC522_Read(8)
                    print ("\n")
                    print ("Sector 8 will now be filled with 0xFF:")
                    # Write the data
                    A.MFRC522_Write(8, data)
                    print ("\n")
                    print ("It now looks like this:")
                    # Check to see if it was written
                    A.MFRC522_Read(8)
                    print ("\n")
                    '''
                    A.MFRC522_StopCrypto1()
                else:
                    print("Authentication error")
        else:
            print('not detected')

# A.reset()



from numpy import int16, std
from PSL.SENSORS.Kalman import KalmanFilter


def connect(route, **args):
    return MPU6050(route, **args)


class MPU6050():
    '''
    Mandatory members:
    GetRaw : Function called by Graphical apps. Must return values stored in a list
    NUMPLOTS : length of list returned by GetRaw. Even single datapoints need to be stored in a list before returning
    PLOTNAMES : a list of strings describing each element in the list returned by GetRaw. len(PLOTNAMES) = NUMPLOTS
    name : the name of the sensor shown to the user
    params:
        A dictionary of function calls(single arguments only) paired with list of valid argument values. (Primitive. I know.)
        These calls can be used for one time configuration settings

    '''
    GYRO_CONFIG = 0x1B
    ACCEL_CONFIG = 0x1C
    GYRO_SCALING = [131, 65.5, 32.8, 16.4]
    ACCEL_SCALING = [16384, 8192, 4096, 2048]
    AR = 3
    GR = 3
    NUMPLOTS = 7
    PLOTNAMES = ['Ax', 'Ay', 'Az', 'Temp', 'Gx', 'Gy', 'Gz']
    ADDRESS = 0x68
    name = 'Accel/gyro'

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.name = 'Accel/gyro'
        self.params = {'powerUp': None, 'setGyroRange': [250, 500, 1000, 2000], 'setAccelRange': [2, 4, 8, 16],
                       'KalmanFilter': {'dataType': 'double', 'min': 0, 'max': 1000, 'prefix': 'value: '}}
        self.setGyroRange(2000)
        self.setAccelRange(16)
        self.powerUp()
        self.K = None

    def KalmanFilter(self, opt):
        if opt == 0:
            self.K = None
            return
        noise = [[]] * self.NUMPLOTS
        for a in range(500):
            vals = self.getRaw()
            for b in range(self.NUMPLOTS): noise[b].append(vals[b])

        self.K = [None] * 7
        for a in range(self.NUMPLOTS):
            sd = std(noise[a])
            self.K[a] = KalmanFilter(1. / opt, sd ** 2)

    def getVals(self, addr, numbytes):
        vals = self.I2C.readBulk(self.ADDRESS, addr, numbytes)
        return vals

    def powerUp(self):
        self.I2C.writeBulk(self.ADDRESS, [0x6B, 0])

    def setGyroRange(self, rs):
        self.GR = self.params['setGyroRange'].index(rs)
        self.I2C.writeBulk(self.ADDRESS, [self.GYRO_CONFIG, self.GR << 3])

    def setAccelRange(self, rs):
        self.AR = self.params['setAccelRange'].index(rs)
        self.I2C.writeBulk(self.ADDRESS, [self.ACCEL_CONFIG, self.AR << 3])

    def getRaw(self):
        '''
        This method must be defined if you want GUIs to use this class to generate
        plots on the fly.
        It must return a set of different values read from the sensor. such as X,Y,Z acceleration.
        The length of this list must not change, and must be defined in the variable NUMPLOTS.

        GUIs will generate as many plots, and the data returned from this method will be appended appropriately
        '''
        vals = self.getVals(0x3B, 14)
        if vals:
            if len(vals) == 14:
                raw = [0] * 7
                for a in range(3): raw[a] = 1. * int16(vals[a * 2] << 8 | vals[a * 2 + 1]) / self.ACCEL_SCALING[self.AR]
                for a in range(4, 7): raw[a] = 1. * int16(vals[a * 2] << 8 | vals[a * 2 + 1]) / self.GYRO_SCALING[
                    self.GR]
                raw[3] = int16(vals[6] << 8 | vals[7]) / 340. + 36.53
                if not self.K:
                    return raw
                else:
                    for b in range(self.NUMPLOTS):
                        self.K[b].input_latest_noisy_measurement(raw[b])
                        raw[b] = self.K[b].get_latest_estimated_measurement()
                    return raw

            else:
                return False
        else:
            return False

    def getAccel(self):
        vals = self.getVals(0x3B, 6)
        ax = int16(vals[0] << 8 | vals[1])
        ay = int16(vals[2] << 8 | vals[3])
        az = int16(vals[4] << 8 | vals[5])
        return [ax / 65535., ay / 65535., az / 65535.]

    def getTemp(self):
        vals = self.getVals(0x41, 6)
        t = int16(vals[0] << 8 | vals[1])
        return t / 65535.

    def getGyro(self):
        vals = self.getVals(0x43, 6)
        ax = int16(vals[0] << 8 | vals[1])
        ay = int16(vals[2] << 8 | vals[3])
        az = int16(vals[4] << 8 | vals[5])
        return [ax / 65535., ay / 65535., az / 65535.]


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    A = connect(I.I2C)
    t, x, y, z = I.I2C.capture(A.ADDRESS, 0x43, 6, 5000, 1000, 'int')
    # print (t,x,y,z)
    from pylab import *

    plot(t, x)
    plot(t, y)
    plot(t, z)
    show()



"""Gas sensors can be used to measure the concentration of certain gases."""

from typing import Callable, Union

from pslab import Multimeter
from pslab.serial_handler import SerialHandler


class MQ135:
    """MQ135 is a cheap gas sensor that can detect several harmful gases.

    The MQ135 is most suitable for detecting flammable gases, but can also be
    used to measure CO2.

    Parameters
    ----------
    gas : {CO2, CO, EtOH, NH3, Tol, Ace}
        The gas to be detected:
            CO2: Carbon dioxide
            CO: Carbon monoxide
            EtOH: Ethanol
            NH3: Ammonia
            Tol: Toluene
            Ace: Acetone
    r_load : float
        Load resistance in ohm.
    device : :class:`SerialHandler`, optional
        Serial connection to PSLab device. If not provided, a new one will be
        created.
    channel : str, optional
        Analog input on which to monitor the sensor output voltage. The default
        value is CH1. Be aware that the sensor output voltage can be as high
        as 5 V, depending on load resistance and gas concentration.
    r0 : float, optional
        The sensor resistance when exposed to 100 ppm NH3 at 20 degC and 65%
        RH. Varies between individual sensors. Optional, but gas concentration
        cannot be measured unless R0 is known. If R0 is not known,
        :meth:`measure_r0` to find it.
    temperature : float or Callable, optional
        Ambient temperature in degC. The default value is 20. A callback can
        be provided in place of a fixed value.
    humidity : float or Callable, optional
        Relative humidity between 0 and 1. The default value is 0.65. A
        callback can be provided in place of a fixed value.
    """

    # Parameters manually extracted from data sheet.
    # ppm = A * (Rs/R0) ^ B
    _PARAMS = {
        "CO2": [109, -2.88],
        "CO": [583, -3.93],
        "EtOH": [76.4, -3.18],
        "NH3": [102, -2.49],
        "Tol": [44.6, -3.45],
        "Ace": [33.9, -3.42],
    }

    # Assuming second degree temperature dependence and linear humidity dependence.
    _TEMPERATURE_CORRECTION = [3.28e-4, -2.55e-2, 1.38]
    _HUMIDITY_CORRECTION = -2.24e-1

    def __init__(
        self,
        gas: str,
        r_load: float,
        device: SerialHandler = None,
        channel: str = "CH1",
        r0: float = None,
        temperature: Union[float, Callable] = 20,
        humidity: Union[float, Callable] = 0.65,
    ):
        self._multimeter = Multimeter(device)
        self._params = self._PARAMS[gas]
        self.channel = channel
        self.r_load = r_load
        self.r0 = r0
        self.vcc = 5

        if isinstance(temperature, Callable):
            self._temperature = temperature
        else:

            def _temperature():
                return temperature

            self._temperature = _temperature

        if isinstance(humidity, Callable):
            self._humidity = humidity
        else:

            def _humidity():
                return humidity

            self._humidity = _humidity

    @property
    def _voltage(self):
        return self._multimeter.measure_voltage(self.channel)

    @property
    def _correction(self):
        """Correct sensor resistance for temperature and humidity.

        Coefficients are averages of curves fitted to temperature data for 33%
        and 85% relative humidity extracted manually from the data sheet.
        Humidity dependence is assumed to be linear, and is centered on 65% RH.
        """
        t = self._temperature()
        h = self._humidity()
        a, b, c, d = *self._TEMPERATURE_CORRECTION, self._HUMIDITY_CORRECTION
        return a * t ** 2 + b * t + c + d * (h - 0.65)

    @property
    def _sensor_resistance(self):
        return (self.vcc / self._voltage - 1) * self.r_load / self._correction

    def measure_concentration(self):
        """Measure the concentration of the configured gas.

        Returns
        -------
        concentration : float
            Gas concentration in ppm.
        """
        try:
            return (
                self._params[0] * (self._sensor_resistance / self.r0) ** self._params[1]
            )
        except TypeError:
            raise TypeError("r0 is not set.")

    def measure_r0(self, gas_concentration: float):
        """Determine sensor resistance at 100 ppm NH3 in otherwise clean air.

        For best results, monitor R0 over several hours and use the average
        value.

        The sensor resistance at 100 ppm NH3 (R0) is used as a reference
        against which the present sensor resistance must be compared in order
        to calculate gas concentration.

        R0 can be determined by calibrating the sensor at any known gas
        concentration.

        Parameters
        ----------
        gas_concentration : float
            A known concentration of the configured gas in ppm.

        Returns
        -------
        r0 : float
            The sensor resistance at 100 ppm NH3 in ohm.
        """
        return self._sensor_resistance * (gas_concentration / self._params[0]) ** (
            1 / -self._params[1]
        )



class KalmanFilter(object):
    '''
    Credits:http://scottlobdell.me/2014/08/kalman-filtering-python-reading-sensor-input/
    '''

    def __init__(self, process_variance, estimated_measurement_variance):
        self.process_variance = process_variance
        self.estimated_measurement_variance = estimated_measurement_variance
        self.posteri_estimate = 0.0
        self.posteri_error_estimate = 1.0

    def input_latest_noisy_measurement(self, measurement):
        priori_estimate = self.posteri_estimate
        priori_error_estimate = self.posteri_error_estimate + self.process_variance

        blending_factor = priori_error_estimate / (priori_error_estimate + self.estimated_measurement_variance)
        self.posteri_estimate = priori_estimate + blending_factor * (measurement - priori_estimate)
        self.posteri_error_estimate = (1 - blending_factor) * priori_error_estimate

    def get_latest_estimated_measurement(self):
        return self.posteri_estimate



"""Low-level communication for PSLab.

Example
-------
>>> from pslab.serial_handler import SerialHandler
>>> device = SerialHandler()
>>> version = device.get_version()
>>> device.disconnect()
"""
import logging
import os.path
import platform
import struct
import time
from functools import partial, update_wrapper
from typing import List, Union

import serial
from serial.tools import list_ports

import pslab.protocol as CP

logger = logging.getLogger(__name__)


def detect():
    """Detect connected PSLab devices.

    Returns
    -------
    devices : dict of str: str
        Dictionary containing port name as keys and device version on that
        port as values.
    """
    regex = []

    for vid, pid in zip(SerialHandler._USB_VID, SerialHandler._USB_PID):
        regex.append(f"{vid:04x}:{pid:04x}")

    regex = "(" + "|".join(regex) + ")"
    port_info_generator = list_ports.grep(regex)
    pslab_devices = {}

    for port_info in port_info_generator:
        version = _get_version(port_info.device)
        if any(expected in version for expected in ["PSLab", "CSpark"]):
            pslab_devices[port_info.device] = version

    return pslab_devices


def _get_version(port: str) -> str:
    interface = serial.Serial(port=port, baudrate=1e6, timeout=1)
    interface.write(CP.COMMON)
    interface.write(CP.GET_VERSION)
    version = interface.readline()
    return version.decode("utf-8")


class SerialHandler:
    """Provides methods for communicating with the PSLab hardware.

    When instantiated, SerialHandler tries to connect to the PSLab. A port can
    optionally be specified; otherwise Handler will try to find the correct
    port automatically.

    Parameters
    ----------
    See :meth:`connect`.
    """

    #            V5      V6
    _USB_VID = [0x04D8, 0x10C4]
    _USB_PID = [0x00DF, 0xEA60]

    def __init__(
        self,
        port: str = None,
        baudrate: int = 1000000,
        timeout: float = 1.0,
    ):
        self._check_udev()
        self.version = ""
        self._log = b""
        self._logging = False
        self.interface = serial.Serial()
        self.send_byte = partial(self._send, size=1)
        update_wrapper(self.send_byte, self._send)
        self.send_int = partial(self._send, size=2)
        update_wrapper(self.send_int, self._send)
        self.get_byte = partial(self._receive, size=1)
        update_wrapper(self.get_byte, self._receive)
        self.get_int = partial(self._receive, size=2)
        update_wrapper(self.get_int, self._receive)
        self.get_long = partial(self._receive, size=4)
        update_wrapper(self.get_long, self._receive)
        self.connect(port=port, baudrate=baudrate, timeout=timeout)
        self.connected = self.interface.is_open

    @staticmethod
    def _check_udev():
        if platform.system() == "Linux":
            udev_paths = [
                "/run/udev/rules.d/",
                "/etc/udev/rules.d/",
                "/lib/udev/rules.d/",
            ]
            for p in udev_paths:
                udev_rules = os.path.join(p, "99-pslab.rules")
                if os.path.isfile(udev_rules):
                    break
            else:
                e = (
                    "A udev rule must be installed to access the PSLab. "
                    + "Please copy 99-pslab.rules to /etc/udev/rules.d/."
                )
                raise OSError(e)

    @staticmethod
    def _list_ports() -> List[str]:
        """Return a list of serial port names."""
        return [p.device for p in list_ports.comports()]

    def connect(
        self,
        port: str = None,
        baudrate: int = 1000000,
        timeout: float = 1.0,
    ):
        """Connect to PSLab.

        Parameters
        ----------
        port : str, optional
            The name of the port to which the PSLab is connected as a string.
            On Posix this is a path, e.g. "/dev/ttyACM0". On Windows, it's a
            numbered COM port, e.g. "COM5". Will be autodetected if not
            specified. If multiple PSLab devices are connected, port must be
            specified.
        baudrate : int, optional
            Symbol rate in bit/s. The default value is 1000000.
        timeout : float, optional
            Time in seconds to wait before cancelling a read or write command. The
            default value is 1.0.

        Raises
        ------
        SerialException
            If connection could not be established.
        RuntimeError
            If ultiple devices are connected and no port was specified.
        """
        # serial.Serial opens automatically if port is not None.
        self.interface = serial.Serial(
            port=port,
            baudrate=baudrate,
            timeout=timeout,
            write_timeout=timeout,
        )
        pslab_devices = detect()

        if self.interface.is_open:
            # User specified a port.
            version = self.get_version()
        else:
            if len(pslab_devices) == 1:
                self.interface.port = list(pslab_devices.keys())[0]
                self.interface.open()
                version = self.get_version()
            elif len(pslab_devices) > 1:
                found = ""

                for port, version in pslab_devices.items():
                    found += f"{port}: {version}"

                raise RuntimeError(
                    "Multiple PSLab devices found:\n"
                    f"{found}"
                    "Please choose a device by specifying a port."
                )
            else:
                version = ""

        if self.interface.port in pslab_devices:
            self.version = version
            logger.info(f"Connected to {self.version} on {self.interface.port}.")
        else:
            self.interface.close()
            self.version = ""
            raise serial.SerialException("Device not found.")

    def disconnect(self):
        """Disconnect from PSLab."""
        self.interface.close()

    def reconnect(
        self,
        port: str = None,
        baudrate: int = None,
        timeout: float = None,
    ):
        """Reconnect to PSLab.

        Will reuse previous settings (port, baudrate, timeout) unless new ones are
        provided.

        Parameters
        ----------
        See :meth:`connect`.
        """
        self.disconnect()

        # Reuse previous settings unless user provided new ones.
        baudrate = self.interface.baudrate if baudrate is None else baudrate
        port = self.interface.port if port is None else port
        timeout = self.interface.timeout if timeout is None else timeout

        self.connect(
            port=port,
            baudrate=baudrate,
            timeout=timeout,
        )

    def get_version(self) -> str:
        """Query PSLab for its version and return it as a decoded string.

        Returns
        -------
        str
            Version string.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.GET_VERSION)
        version = self.interface.readline()
        self._write_log(version, "RX")
        return version.decode("utf-8")

    def get_ack(self) -> int:
        """Get response code from PSLab.

        Returns
        -------
        int
            Response code. Meanings:
                0x01 ACK
                0x10 I2C ACK
                0x20 I2C bus collision
                0x10 Radio max retransmits
                0x20 Radio not present
                0x40 Radio reply timout
        """
        response = self.read(1)

        if not response:
            raise serial.SerialException("Timeout while waiting for ACK.")

        ack = CP.Byte.unpack(response)[0]

        if not (ack & 0x01):
            raise serial.SerialException("Received non ACK byte while waiting for ACK.")

        return ack

    @staticmethod
    def _get_integer_type(size: int) -> struct.Struct:
        if size == 1:
            return CP.Byte
        elif size == 2:
            return CP.ShortInt
        elif size == 4:
            return CP.Integer
        else:
            raise ValueError("size must be 1, 2, or 4.")

    def _send(self, value: Union[bytes, int], size: int):
        """Send a value to the PSLab.

        Parameters
        ----------
        value : int
            Value to send to PSLab. Must fit in four bytes.
        """
        if isinstance(value, bytes):
            packet = value
        else:
            packer = self._get_integer_type(size)
            packet = packer.pack(value)

        self.write(packet)

    def _receive(self, size: int) -> int:
        """Read and unpack data from the serial port.

        Returns
        -------
        int
            Unpacked data.

        Raises
        ------
        SerialException if too few bytes received.
        """
        received = self.read(size)

        if len(received) == size:
            unpacker = self._get_integer_type(size)
            retval = unpacker.unpack(received)[0]
        else:
            raise serial.SerialException(
                f"Requested {size} bytes, got {len(received)}."
            )

        return retval

    def read(self, number_of_bytes: int) -> bytes:
        """Log incoming bytes.

        Wrapper for Serial.read().

        Parameters
        ----------
        number_of_bytes : int
            Number of bytes to read from the serial port.

        Returns
        -------
        bytes
            Bytes read from the serial port.
        """
        data = self.interface.read(number_of_bytes)
        self._write_log(data, "RX")
        return data

    def write(self, data: bytes):
        """Log outgoing bytes.

        Wrapper for Serial.write().

        Parameters
        ----------
        data : int
            Bytes to write to the serial port.
        """
        self.interface.write(data)
        self._write_log(data, "TX")

    def _write_log(self, data: bytes, direction: str):
        if self._logging:
            self._log += direction.encode() + data + "STOP".encode()

    def wait_for_data(self, timeout: float = 0.2) -> bool:
        """Wait for :timeout: seconds or until there is data in the input buffer.

        Parameters
        ----------
        timeout : float, optional
            Time in seconds to wait. The default is 0.2.

        Returns
        -------
        bool
            True iff the input buffer is not empty.
        """
        start_time = time.time()

        while time.time() - start_time < timeout:
            if self.interface.in_waiting:
                return True
            time.sleep(0.02)

        return False


RECORDED_TRAFFIC = iter([])
"""An iterator returning (request, response) pairs.

The request is checked against data written to the dummy serial port, and if it matches
the response can be read back. Both request and response should be bytes-like.

Intended to be monkey-patched by the calling test module.
"""


class MockHandler(SerialHandler):
    """Mock implementation of :class:`SerialHandler` for testing.

    Parameters
    ----------
    Same as :class:`SerialHandler`.
    """

    VERSION = "PSLab vMOCK"

    def __init__(
        self,
        port: str = None,
        baudrate: int = 1000000,
        timeout: float = 1.0,
    ):
        self._in_buffer = b""
        super().__init__(port, baudrate, timeout)

    @staticmethod
    def _check_udev():
        pass

    def connect(
        self,
        port: str = None,
        baudrate: int = 1000000,
        timeout: float = 1.0,
    ):
        """See :meth:`SerialHandler.connect`."""
        self.version = self.get_version()

    def disconnect(self):
        """See :meth:`SerialHandler.disconnect`."""
        pass

    def reconnect(
        self,
        port: str = None,
        baudrate: int = None,
        timeout: float = None,
    ):
        """See :meth:`SerialHandler.reconnect`."""
        pass

    def get_version(self) -> str:
        """Return mock version."""
        return self.VERSION

    def read(self, number_of_bytes: int) -> bytes:
        """Mimic the behavior of the serial bus by returning recorded RX traffic.

        The returned data depends on how :meth:`write` was called prior to calling
        :meth:`read`.

        See also :meth:`SerialHandler.read`.
        """
        read_bytes = self._in_buffer[:number_of_bytes]
        self._in_buffer = self._in_buffer[number_of_bytes:]
        return read_bytes

    def write(self, data: bytes):
        """Add recorded RX data to buffer if written data equals recorded TX data.

        See also :meth:`SerialHandler.write`.
        """
        tx, rx = next(RECORDED_TRAFFIC)
        if tx == data:
            self._in_buffer += rx

    def wait_for_data(self, timeout: float = 0.2) -> bool:
        """Return True if there is data in buffer, or return False after timeout."""
        if self._in_buffer:
            return True
        else:
            time.sleep(timeout)
            return False


class ADCBufferMixin:
    """Mixin for classes that need to read or write to the ADC buffer."""

    def fetch_buffer(self, samples: int, starting_position: int = 0):
        """Fetch a section of the ADC buffer.

        Parameters
        ----------
        samples : int
            Number of samples to fetch.
        starting_position : int, optional
            Location in the ADC buffer to start from. By default samples will
            be fetched from the beginning of the buffer.

        Returns
        -------
        received : list of int
            List of received samples.
        """
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.RETRIEVE_BUFFER)
        self._device.send_int(starting_position)
        self._device.send_int(samples)
        received = [self._device.get_int() for i in range(samples)]
        self._device.get_ack()
        return received

    def clear_buffer(self, samples: int, starting_position: int = 0):
        """Clear a section of the ADC buffer.

        Parameters
        ----------
        samples : int
            Number of samples to clear from the buffer.
        starting_position : int, optional
            Location in the ADC buffer to start from. By default samples will
            be cleared from the beginning of the buffer.
        """
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.CLEAR_BUFFER)
        self._device.send_int(starting_position)
        self._device.send_int(samples)
        self._device.get_ack()

    def fill_buffer(self, data: List[int], starting_position: int = 0):
        """Fill a section of the ADC buffer with data.

        Parameters
        ----------
        data : list of int
            Values to write to the ADC buffer.
        starting_position : int, optional
            Location in the ADC buffer to start from. By default writing will
            start at the beginning of the buffer.
        """
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.FILL_BUFFER)
        self._device.send_int(starting_position)
        self._device.send_int(len(data))

        for value in data:
            self._device.send_int(value)

        self._device.get_ack()



from __future__ import print_function

import time

from numpy import int16


def connect(route, **args):
    return BMP180(route, **args)


class BMP180:
    ADDRESS = 0x77
    REG_CONTROL = 0xF4
    REG_RESULT = 0xF6
    CMD_TEMP = 0x2E
    CMD_P0 = 0x34
    CMD_P1 = 0x74
    CMD_P2 = 0xB4
    CMD_P3 = 0xF4
    oversampling = 0
    NUMPLOTS = 3
    PLOTNAMES = ['Temperature', 'Pressure', 'Altitude']
    name = 'Altimeter BMP180'

    def __init__(self, I2C, **args):
        self.ADDRESS = args.get('address', self.ADDRESS)

        self.I2C = I2C

        self.MB = self.__readInt__(0xBA)

        self.c3 = 160.0 * pow(2, -15) * self.__readInt__(0xAE)
        self.c4 = pow(10, -3) * pow(2, -15) * self.__readUInt__(0xB0)
        self.b1 = pow(160, 2) * pow(2, -30) * self.__readInt__(0xB6)
        self.c5 = (pow(2, -15) / 160) * self.__readUInt__(0xB2)
        self.c6 = self.__readUInt__(0xB4)
        self.mc = (pow(2, 11) / pow(160, 2)) * self.__readInt__(0xBC)
        self.md = self.__readInt__(0xBE) / 160.0
        self.x0 = self.__readInt__(0xAA)
        self.x1 = 160.0 * pow(2, -13) * self.__readInt__(0xAC)
        self.x2 = pow(160, 2) * pow(2, -25) * self.__readInt__(0xB8)
        self.y0 = self.c4 * pow(2, 15)
        self.y1 = self.c4 * self.c3
        self.y2 = self.c4 * self.b1
        self.p0 = (3791.0 - 8.0) / 1600.0
        self.p1 = 1.0 - 7357.0 * pow(2, -20)
        self.p2 = 3038.0 * 100.0 * pow(2, -36)
        self.T = 25
        print('calib:', self.c3, self.c4, self.b1, self.c5, self.c6, self.mc, self.md, self.x0, self.x1, self.x2,
              self.y0, self.y1, self.p0, self.p1, self.p2)
        self.params = {'setOversampling': [0, 1, 2, 3]}
        self.name = 'BMP180 Altimeter'
        self.initTemperature()
        self.readTemperature()
        self.initPressure()
        self.baseline = self.readPressure()

    def __readInt__(self, addr):
        return int16(self.__readUInt__(addr))

    def __readUInt__(self, addr):
        vals = self.I2C.readBulk(self.ADDRESS, addr, 2)
        v = 1. * ((vals[0] << 8) | vals[1])
        return v

    def initTemperature(self):
        self.I2C.writeBulk(self.ADDRESS, [self.REG_CONTROL, self.CMD_TEMP])
        time.sleep(0.005)

    def readTemperature(self):
        vals = self.I2C.readBulk(self.ADDRESS, self.REG_RESULT, 2)
        if len(vals) == 2:
            T = (vals[0] << 8) + vals[1]
            a = self.c5 * (T - self.c6)
            self.T = a + (self.mc / (a + self.md))
            return self.T
        else:
            return False

    def setOversampling(self, num):
        self.oversampling = num

    def initPressure(self):
        os = [0x34, 0x74, 0xb4, 0xf4]
        delays = [0.005, 0.008, 0.014, 0.026]
        self.I2C.writeBulk(self.ADDRESS, [self.REG_CONTROL, os[self.oversampling]])
        time.sleep(delays[self.oversampling])

    def readPressure(self):
        vals = self.I2C.readBulk(self.ADDRESS, self.REG_RESULT, 3)
        if len(vals) == 3:
            P = 1. * (vals[0] << 8) + vals[1] + (vals[2] / 256.0)
            s = self.T - 25.0
            x = (self.x2 * pow(s, 2)) + (self.x1 * s) + self.x0
            y = (self.y2 * pow(s, 2)) + (self.y1 * s) + self.y0
            z = (P - x) / y
            self.P = (self.p2 * pow(z, 2)) + (self.p1 * z) + self.p0
            return self.P
        else:
            return False

    def altitude(self):
        # baseline pressure needs to be provided
        return (44330.0 * (1 - pow(self.P / self.baseline, 1 / 5.255)))

    def sealevel(self, P, A):
        '''
        given a calculated pressure and altitude, return the sealevel
        '''
        return (P / pow(1 - (A / 44330.0), 5.255))

    def getRaw(self):
        self.initTemperature()
        self.readTemperature()
        self.initPressure()
        self.readPressure()
        return [self.T, self.P, self.altitude()]



"""Functions related to CLI for PSLab.

Example
-------
>>> from pslab import cli
>>> parser, subparser = cli.get_parser()
>>> cli.add_collect_args(subparser)
>>> cli.add_wave_args(subparser)
>>> cli.add_pwm_args(subparser)
>>> parser.parse_args(["collect","-i","logic_analyzer"])
Namespace(channels=1, duration=1, file_path=None, function='collect',
instrument='logic_analyzer', json=False, port=None)
"""

import argparse
import csv
import json
import sys
import time
from itertools import zip_longest
from typing import List, Tuple

import numpy as np

import pslab.protocol as CP
from pslab.instrument.logic_analyzer import LogicAnalyzer
from pslab.instrument.oscilloscope import Oscilloscope
from pslab.instrument.waveform_generator import WaveformGenerator, PWMGenerator
from pslab.serial_handler import SerialHandler


def logic_analyzer(
    device: SerialHandler, channels: int, duration: float
) -> Tuple[List[str], List[np.ndarray]]:
    """Capture logic events on up to four channels simultaneously.

    Parameters
    ----------
    device : :class:`Handler`
        Serial interface for communicating with the PSLab device.
    channels : {1, 2, 3, 4}
        Number of channels to capture events on. Events will be captured on LA1,
        LA2, LA3, and LA4, in that order.
    duration : float
        Duration in seconds up to which events will be captured.

    Returns
    -------
    list of str
        Name of active channels.
    list of numpy.ndarray
        List of numpy.ndarrays holding timestamps in microseconds when logic events
        were detected. The length of the list is equal to the number of channels
        that were used to capture events, and each list element corresponds to a
        channel.

    Warnings
    --------
        This cannot be used at the same time as the oscilloscope.
    """
    la = LogicAnalyzer(device)
    la.capture(channels, block=False)
    time.sleep(duration)
    la.stop()
    timestamps = la.fetch_data()
    channels_name = [la._channel_one_map, la._channel_two_map, "LA3", "LA4"]

    return channels_name[:channels], timestamps


def oscilloscope(
    device: SerialHandler, channels: int, duration: float
) -> Tuple[List[str], List[np.ndarray]]:
    """Capture varying voltage signals on up to four channels simultaneously.

    Parameters
    ----------
    device : :class:`Handler`
        Serial interface for communicating with the PSLab device.
    channels : {1, 2, 4}
        Number of channels to sample from simultaneously. By default, samples are
        captured from CH1, CH2, CH3 and MIC.
    duration : float
        Duration in seconds up to which samples will be captured.

    Returns
    -------
    list of str
        "Timestamp", Name of active channels.
    list of numpy.ndarray
        List of numpy.ndarrays with timestamps in the first index and corresponding
        voltages in the following index. The length of the list is equal to one
        additional to the number of channels that were used to capture samples.
    """
    scope = Oscilloscope(device)
    max_samples = CP.MAX_SAMPLES // channels
    min_timegap = scope._lookup_mininum_timegap(channels)
    max_duration = max_samples * min_timegap * 1e-6
    active_channels = ([scope._channel_one_map] + scope._CH234)[:channels]
    xy = [np.array([]) for _ in range(1 + channels)]

    while duration > 0:
        if duration >= max_duration:
            samples = max_samples
        else:
            samples = round((duration * 1e6) / min_timegap)

        st = time.time()
        xy = np.append(xy, scope.capture(channels, samples, min_timegap), axis=1)
        duration -= time.time() - st

    return ["Timestamp"] + active_channels, xy


INSTRUMENTS = {
    "logic_analyzer": logic_analyzer,
    "oscilloscope": oscilloscope,
}


def collect(handler: SerialHandler, args: argparse.Namespace):
    """Collect data from instruments, and write it in file or stdout.

    Parameters
    ----------
    handler : :class:`Handler`
        Serial interface for communicating with the PSLab device.
    args : :class:`argparse.Namespace`
        Parsed arguments.

    Raises
    ------
    LookupError
        If the given instrument not available.
    """
    instrument = INSTRUMENTS.get(args.instrument)

    if instrument is None:
        raise LookupError(args.instrument + " not available")

    output = instrument(handler, args.channels, args.duration)

    if args.file_path is not None:
        file = open(args.file_path, "w")
    else:
        file = sys.stdout

    if not args.json:
        csv_file = csv.writer(file)
        csv_file.writerow(output[0])
        for row in zip_longest(*output[1]):
            csv_file.writerow(row)
    else:
        output_dict = dict()
        for key, val in zip_longest(*output):
            output_dict[key] = val.tolist()
        json.dump(output_dict, file)

    if args.file_path is not None:
        file.close()


def wave(handler: SerialHandler, args: argparse.Namespace):
    """Generate or load wave.

    Parameters
    ----------
    handler : :class:`Handler`
        Serial interface for communicating with the PSLab device.
    args : :class:`argparse.Namespace`
        Parsed arguments.
    """
    waveform_generator = WaveformGenerator(handler)

    if args.wave_function == "gen":
        waveform_generator.generate(
            channels=args.channel,
            frequency=args.frequency,
            phase=args.phase,
        )
    elif args.wave_function == "load":
        if args.table is not None:
            table = args.table
        elif args.table_file is not None:
            with open(args.table_file) as table_file:
                table = json.load(table_file)

        x = np.arange(0, len(table), len(table) / 512)
        y = [table[int(i)] for i in x]
        waveform_generator.load_table(channel=args.channel, points=y)


def pwm(handler: SerialHandler, args: argparse.Namespace):
    """Generate PWM.

    Parameters
    ----------
    handler : :class:`Handler`
        Serial interface for communicating with the PSLab device.
    args : :class:`argparse.Namespace`
        Parsed arguments.
    """
    pwm_generator = PWMGenerator(handler)

    if args.pwm_function == "gen":
        pwm_generator.generate(
            channels=args.channel,
            frequency=args.frequency,
            duty_cycles=args.duty_cycles,
            phases=args.phases,
        )
    elif args.pwm_function == "map":
        pwm_generator.map_reference_clock(
            channels=args.channel,
            prescaler=args.prescaler,
        )


def main(args: argparse.Namespace):
    """Perform the given function on PSLab.

    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Parsed arguments.
    """
    handler = SerialHandler(port=args.port)

    if args.function == "collect":
        collect(handler, args)
    elif args.function == "wave":
        wave(handler, args)
    elif args.function == "pwm":
        pwm(handler, args)


def get_parser() -> Tuple[argparse.ArgumentParser, argparse._SubParsersAction]:
    """Parser for CLI.

    Returns
    -------
    parser : :class:`argparse.ArgumentParser`
        Arqument parser for CLI.
    functions : :class:`argparse._SubParsersAction`
        SubParser to add other arguments related to different function.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--port",
        type=str,
        default=None,
        required=False,
        help="The name of the port to which the PSLab is connected",
    )
    functions = parser.add_subparsers(
        title="Functions", dest="function", description="Functions to perform on PSLab."
    )

    return parser, functions


def add_collect_args(subparser: argparse._SubParsersAction):
    """Add arguments for collect function to ArgumentParser.

    Parameters
    ----------
    subparser : :class:`argparse._SubParsersAction`
        SubParser to add other arguments related to collect function.
    """
    description = "Available Instruments: " + ", ".join(INSTRUMENTS) + "."
    collect = subparser.add_parser("collect", description=description)
    collect.add_argument(
        "instrument",
        type=str,
        help="The name of the instrument to use",
    )
    collect.add_argument(
        "-c",
        "--channels",
        type=int,
        default=1,
        required=False,
        help="Number of channels to capture",
    )
    collect.add_argument(
        "-d",
        "--duration",
        type=float,
        default=1,
        required=False,
        help="Duration for capturing (in seconds)",
    )
    collect.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        required=False,
        dest="file_path",
        help="File name to write data, otherwise in stdout",
    )
    collect.add_argument(
        "-j",
        "--json",
        action="store_true",
        default=False,
        help="Enable it to write data in json format",
    )


def add_wave_args(subparser: argparse._SubParsersAction):
    """Add arguments for wave {gen,load} function to ArgumentParser.

    Parameters
    ----------
    subparser : :class:`argparse._SubParsersAction`
        SubParser to add other arguments related to wave_gen function.
    """
    wave = subparser.add_parser("wave")
    wave_functions = wave.add_subparsers(
        title="Wave Functions",
        dest="wave_function",
    )
    wave_gen = wave_functions.add_parser("gen")
    wave_gen.add_argument(
        "channel",
        nargs="+",
        choices=["SI1", "SI2"],
        help="Pin(s) on which to generate a waveform",
    )
    wave_gen.add_argument(
        "-f",
        "--frequency",
        nargs="+",
        type=float,
        required=True,
        help="Frequency in Hz",
    )
    wave_gen.add_argument(
        "-p",
        "--phase",
        type=float,
        default=0,
        required=False,
        help="Phase between waveforms in degrees",
    )
    description = """
        TABLE:
            JSON array of voltage values which make up the waveform. Array length
            must be 512. If the array length less than 512, then the array will be
            expanded in length of 512. Values outside the range -3.3 V to 3.3 V
            will be clipped.

            examples:
                [1,0] or [1,...,0,...],
                [0,1,0,-1,0,1,0,-1,...],
                [0,.025,.05,.075,.1,.125,.15,...]
    """
    load = wave_functions.add_parser("load", description=description)
    load.add_argument(
        "channel",
        choices=["SI1", "SI2"],
        help="Pin(s) on which to load a table",
    )
    load_table = load.add_mutually_exclusive_group(required=True)
    load_table.add_argument(
        "--table",
        type=json.loads,
        default=None,
        help="Table to load in pin SI1 as json",
    )
    load_table.add_argument(
        "--table-file",
        nargs="?",
        type=str,
        const=0,
        default=None,
        help="Table to load in pin SI1 as json file. Default is stdin",
    )


def add_pwm_args(subparser: argparse._SubParsersAction):
    """Add arguments for pwm {gen,map,set} function to ArgumentParser.

    Parameters
    ----------
    subparser : :class:`argparse._SubParsersAction`
        SubParser to add other arguments related to pwm_gen function.
    """
    pwm = subparser.add_parser("pwm")
    pwm_functions = pwm.add_subparsers(
        title="PWM Functions",
        dest="pwm_function",
    )
    pwm_gen = pwm_functions.add_parser("gen")
    pwm_gen.add_argument(
        "channel",
        nargs="+",
        choices=["SQ1", "SQ2", "SQ3", "SQ4"],
        help="Pin(s) on which to generate a PWM signals",
    )
    pwm_gen.add_argument(
        "-f",
        "--frequency",
        type=float,
        required=True,
        help="Frequency in Hz. Shared by all outputs",
    )
    pwm_gen.add_argument(
        "-d",
        "--duty-cycles",
        nargs="+",
        type=float,
        required=True,
        help="Duty cycle between 0 and 1",
    )
    pwm_gen.add_argument(
        "-p",
        "--phases",
        nargs="+",
        type=float,
        default=0,
        required=False,
        help="Phase between 0 and 1",
    )
    map_ = pwm_functions.add_parser("map")
    map_.add_argument(
        "channel",
        nargs="+",
        choices=["SQ1", "SQ2", "SQ3", "SQ4"],
        help="Digital output pin(s) to which to map the internal oscillator",
    )
    map_.add_argument(
        "-p",
        "--prescaler",
        type=int,
        required=True,
        help="Prescaler value in interval [0, 15]."
        + "The output frequency is 128 / (1 << prescaler) MHz",
    )


def cmdline(args: List[str] = None):
    """Command line for pslab.

    Parameters
    ----------
    args : list of strings.
        Arguments to parse.
    """
    if args is None:
        args = sys.argv[1:]

    parser, subparser = get_parser()
    add_collect_args(subparser)
    add_wave_args(subparser)
    add_pwm_args(subparser)
    main(parser.parse_args(args))



"""Motor control related classes.

Examples
--------
>>> from pslab.external.motor import Servo
>>> servo = Servo("SQ1")
>>> servo.angle = 30  # Turn motor to 30 degrees position.
"""
from typing import Union

from pslab.instrument.waveform_generator import PWMGenerator

MICROSECONDS = 1e6


class Servo:
    """Control servo motors on SQ1-4.

    Parameters
    ----------
    pin : {"SQ1", "SQ2", "SQ3", "SQ4"}
        Name of the digital output on which to generate the control signal.
    pwm_generator : :class:`PWMGenerator`, optional
        PWMGenerator instance with which to generate the control signal.
        Created automatically if not specified. When contolling multiple
        servos, they should all use the same PWMGenerator instance.
    min_angle_pulse : int, optional
        Pulse length in microseconds corresponding to the minimum (0 degree)
        angle of the servo. The default value is 500.
    max_angle_pulse : int, optional
        Pulse length in microseconds corresponding to the maximum (180 degree)
        angle of the servo. The default value is 2500.
    angle_range : int
        Range of the servo in degrees. The default value is 180.
    frequency : float, optional
        Frequency of the control signal in Hz. The default value is 50.
    """

    def __init__(
        self,
        pin: str,
        pwm_generator: PWMGenerator = None,
        min_angle_pulse: int = 500,
        max_angle_pulse: int = 2500,
        angle_range: int = 180,
        frequency: float = 50,
    ):
        self._pwm = PWMGenerator() if pwm_generator is None else pwm_generator
        self._pin = pin
        self._angle = None
        self._min_angle_pulse = min_angle_pulse
        self._max_angle_pulse = max_angle_pulse
        self._angle_range = angle_range
        self._frequency = frequency

    @property
    def angle(self) -> Union[int, None]:
        """:obj:`int` or :obj:`None`: Angle of the servo in degrees."""
        return self._angle

    @angle.setter
    def angle(self, value: int):
        duty_cycle = self._get_duty_cycle(value)
        self._pwm.generate(self._pin, self._frequency, duty_cycle)
        self._angle = value

    def _get_duty_cycle(self, angle):
        angle /= self._angle_range  # Normalize
        angle *= self._max_angle_pulse - self._min_angle_pulse  # Scale
        angle += self._min_angle_pulse  # Offset
        return angle / (self._frequency ** -1 * MICROSECONDS)



# -*- coding: utf-8; mode: python; indent-tabs-mode: t; tab-width:4 -*-
from __future__ import print_function

import time

from numpy import int16

try:
    from collections import OrderedDict
except ImportError:
    # fallback: try to use the ordereddict backport when using python 2.6
    from ordereddict import OrderedDict


def connect(route, **args):
    return ADS1115(route, **args)


class ADS1115:
    ADDRESS = 0x48  # addr pin grounded. floating

    REG_POINTER_MASK = 0x3
    REG_POINTER_CONVERT = 0
    REG_POINTER_CONFIG = 1
    REG_POINTER_LOWTHRESH = 2
    REG_POINTER_HITHRESH = 3

    REG_CONFIG_OS_MASK = 0x8000
    REG_CONFIG_OS_SINGLE = 0x8000
    REG_CONFIG_OS_BUSY = 0x0000
    REG_CONFIG_OS_NOTBUSY = 0x8000

    REG_CONFIG_MUX_MASK = 0x7000
    REG_CONFIG_MUX_DIFF_0_1 = 0x0000  # Differential P = AIN0, N = AIN1 =default)
    REG_CONFIG_MUX_DIFF_0_3 = 0x1000  # Differential P = AIN0, N = AIN3
    REG_CONFIG_MUX_DIFF_1_3 = 0x2000  # Differential P = AIN1, N = AIN3
    REG_CONFIG_MUX_DIFF_2_3 = 0x3000  # Differential P = AIN2, N = AIN3
    REG_CONFIG_MUX_SINGLE_0 = 0x4000  # Single-ended AIN0
    REG_CONFIG_MUX_SINGLE_1 = 0x5000  # Single-ended AIN1
    REG_CONFIG_MUX_SINGLE_2 = 0x6000  # Single-ended AIN2
    REG_CONFIG_MUX_SINGLE_3 = 0x7000  # Single-ended AIN3

    REG_CONFIG_PGA_MASK = 0x0E00  # bits 11:9
    REG_CONFIG_PGA_6_144V = (0 << 9)  # +/-6.144V range = Gain 2/3
    REG_CONFIG_PGA_4_096V = (1 << 9)  # +/-4.096V range = Gain 1
    REG_CONFIG_PGA_2_048V = (2 << 9)  # +/-2.048V range = Gain 2 =default)
    REG_CONFIG_PGA_1_024V = (3 << 9)  # +/-1.024V range = Gain 4
    REG_CONFIG_PGA_0_512V = (4 << 9)  # +/-0.512V range = Gain 8
    REG_CONFIG_PGA_0_256V = (5 << 9)  # +/-0.256V range = Gain 16

    REG_CONFIG_MODE_MASK = 0x0100  # bit 8
    REG_CONFIG_MODE_CONTIN = (0 << 8)  # Continuous conversion mode
    REG_CONFIG_MODE_SINGLE = (1 << 8)  # Power-down single-shot mode =default)

    REG_CONFIG_DR_MASK = 0x00E0
    REG_CONFIG_DR_8SPS = (0 << 5)  # 8 SPS
    REG_CONFIG_DR_16SPS = (1 << 5)  # 16 SPS
    REG_CONFIG_DR_32SPS = (2 << 5)  # 32 SPS
    REG_CONFIG_DR_64SPS = (3 << 5)  # 64 SPS
    REG_CONFIG_DR_128SPS = (4 << 5)  # 128 SPS
    REG_CONFIG_DR_250SPS = (5 << 5)  # 260 SPS
    REG_CONFIG_DR_475SPS = (6 << 5)  # 475 SPS
    REG_CONFIG_DR_860SPS = (7 << 5)  # 860 SPS

    REG_CONFIG_CMODE_MASK = 0x0010
    REG_CONFIG_CMODE_TRAD = 0x0000
    REG_CONFIG_CMODE_WINDOW = 0x0010

    REG_CONFIG_CPOL_MASK = 0x0008
    REG_CONFIG_CPOL_ACTVLOW = 0x0000
    REG_CONFIG_CPOL_ACTVHI = 0x0008

    REG_CONFIG_CLAT_MASK = 0x0004
    REG_CONFIG_CLAT_NONLAT = 0x0000
    REG_CONFIG_CLAT_LATCH = 0x0004

    REG_CONFIG_CQUE_MASK = 0x0003
    REG_CONFIG_CQUE_1CONV = 0x0000
    REG_CONFIG_CQUE_2CONV = 0x0001
    REG_CONFIG_CQUE_4CONV = 0x0002
    REG_CONFIG_CQUE_NONE = 0x0003
    gains = OrderedDict([('GAIN_TWOTHIRDS', REG_CONFIG_PGA_6_144V), ('GAIN_ONE', REG_CONFIG_PGA_4_096V),
                         ('GAIN_TWO', REG_CONFIG_PGA_2_048V), ('GAIN_FOUR', REG_CONFIG_PGA_1_024V),
                         ('GAIN_EIGHT', REG_CONFIG_PGA_0_512V), ('GAIN_SIXTEEN', REG_CONFIG_PGA_0_256V)])
    gain_scaling = OrderedDict(
        [('GAIN_TWOTHIRDS', 0.1875), ('GAIN_ONE', 0.125), ('GAIN_TWO', 0.0625), ('GAIN_FOUR', 0.03125),
         ('GAIN_EIGHT', 0.015625), ('GAIN_SIXTEEN', 0.0078125)])
    type_selection = OrderedDict(
        [('UNI_0', 0), ('UNI_1', 1), ('UNI_2', 2), ('UNI_3', 3), ('DIFF_01', '01'), ('DIFF_23', '23')])
    sdr_selection = OrderedDict(
        [(8, REG_CONFIG_DR_8SPS), (16, REG_CONFIG_DR_16SPS), (32, REG_CONFIG_DR_32SPS), (64, REG_CONFIG_DR_64SPS),
         (128, REG_CONFIG_DR_128SPS), (250, REG_CONFIG_DR_250SPS), (475, REG_CONFIG_DR_475SPS),
         (860, REG_CONFIG_DR_860SPS)])  # sampling data rate

    NUMPLOTS = 1
    PLOTNAMES = ['mV']

    def __init__(self, I2C, **args):
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.I2C = I2C
        self.channel = 'UNI_0'
        self.gain = 'GAIN_ONE'
        self.rate = 128

        self.setGain('GAIN_ONE')
        self.setChannel('UNI_0')
        self.setDataRate(128)
        self.conversionDelay = 8
        self.name = 'ADS1115 16-bit ADC'
        self.params = {'setGain': self.gains.keys(), 'setChannel': self.type_selection.keys(),
                       'setDataRate': self.sdr_selection.keys()}

    def __readInt__(self, addr):
        return int16(self.__readUInt__(addr))

    def __readUInt__(self, addr):
        vals = self.I2C.readBulk(self.ADDRESS, addr, 2)
        v = 1. * ((vals[0] << 8) | vals[1])
        return v

    def initTemperature(self):
        self.I2C.writeBulk(self.ADDRESS, [self.REG_CONTROL, self.CMD_TEMP])
        time.sleep(0.005)

    def readRegister(self, register):
        vals = self.I2C.readBulk(self.ADDRESS, register, 2)
        return (vals[0] << 8) | vals[1]

    def writeRegister(self, reg, value):
        self.I2C.writeBulk(self.ADDRESS, [reg, (value >> 8) & 0xFF, value & 0xFF])

    def setGain(self, gain):
        '''
        options : 'GAIN_TWOTHIRDS','GAIN_ONE','GAIN_TWO','GAIN_FOUR','GAIN_EIGHT','GAIN_SIXTEEN'
        '''
        self.gain = gain

    def setChannel(self, channel):
        '''
        options 'UNI_0','UNI_1','UNI_2','UNI_3','DIFF_01','DIFF_23'
        '''
        self.channel = channel

    def setDataRate(self, rate):
        '''
        data rate options 8,16,32,64,128,250,475,860 SPS
        '''
        self.rate = rate

    def readADC_SingleEnded(self, chan):
        if chan > 3: return None
        # start with default values
        config = (self.REG_CONFIG_CQUE_NONE  # Disable the comparator (default val)
                  | self.REG_CONFIG_CLAT_NONLAT  # Non-latching (default val)
                  | self.REG_CONFIG_CPOL_ACTVLOW  # Alert/Rdy active low   (default val)
                  | self.REG_CONFIG_CMODE_TRAD  # Traditional comparator (default val)
                  | self.sdr_selection[self.rate]  # 1600 samples per second (default)
                  | self.REG_CONFIG_MODE_SINGLE)  # Single-shot mode (default)

        # Set PGA/voltage range
        config |= self.gains[self.gain]

        if chan == 0:
            config |= self.REG_CONFIG_MUX_SINGLE_0
        elif chan == 1:
            config |= self.REG_CONFIG_MUX_SINGLE_1
        elif chan == 2:
            config |= self.REG_CONFIG_MUX_SINGLE_2
        elif chan == 3:
            config |= self.REG_CONFIG_MUX_SINGLE_3
        # Set 'start single-conversion' bit
        config |= self.REG_CONFIG_OS_SINGLE
        self.writeRegister(self.REG_POINTER_CONFIG, config);
        time.sleep(1. / self.rate + .002)  # convert to mS to S
        return self.readRegister(self.REG_POINTER_CONVERT) * self.gain_scaling[self.gain]

    def readADC_Differential(self, chan='01'):
        # start with default values
        config = (self.REG_CONFIG_CQUE_NONE  # Disable the comparator (default val)
                  | self.REG_CONFIG_CLAT_NONLAT  # Non-latching (default val)
                  | self.REG_CONFIG_CPOL_ACTVLOW  # Alert/Rdy active low   (default val)
                  | self.REG_CONFIG_CMODE_TRAD  # Traditional comparator (default val)
                  | self.sdr_selection[self.rate]  # samples per second
                  | self.REG_CONFIG_MODE_SINGLE)  # Single-shot mode (default)

        # Set PGA/voltage range
        config |= self.gains[self.gain]
        if chan == '01':
            config |= self.REG_CONFIG_MUX_DIFF_0_1
        elif chan == '23':
            config |= self.REG_CONFIG_MUX_DIFF_2_3
        # Set 'start single-conversion' bit
        config |= self.REG_CONFIG_OS_SINGLE
        self.writeRegister(self.REG_POINTER_CONFIG, config);
        time.sleep(1. / self.rate + .002)  # convert to mS to S
        return int16(self.readRegister(self.REG_POINTER_CONVERT)) * self.gain_scaling[self.gain]

    def getLastResults(self):
        return int16(self.readRegister(self.REG_POINTER_CONVERT)) * self.gain_scaling[self.gain]

    def getRaw(self):
        '''
        return values in mV
        '''
        chan = self.type_selection[self.channel]
        if self.channel[:3] == 'UNI':
            return [self.readADC_SingleEnded(chan)]
        elif self.channel[:3] == 'DIF':
            return [self.readADC_Differential(chan)]



from __future__ import print_function

import time

import numpy as np

'''
The running directory must have caldata.py containing dictionary called 'calibs' in it.
The entries of the dictionary will be of the form
calibs={
'AIN6AINCOM':[6.993123e-07,-1.563294e-06,9.994211e-01,-4.596018e-03], 
...
}

'''


def _bv(x):
    return 1 << x


class AD7718:
    VREF = 3.3

    STATUS = 0
    MODE = 1
    ADCCON = 2
    FILTER = 3
    ADCDATA = 4
    ADCOFFSET = 5
    ADCGAIN = 6
    IOCON = 7
    TEST1 = 12
    TEST2 = 13
    ID = 15
    # bit definitions
    MODE_PD = 0
    MODE_IDLE = 1
    MODE_SINGLE = 2
    MODE_CONT = 3
    MODE_INT_ZEROCAL = 4
    MODE_INT_FULLCAL = 5
    MODE_SYST_ZEROCAL = 6
    MODE_SYST_FULLCAL = 7

    MODE_OSCPD = _bv(3)
    MODE_CHCON = _bv(4)
    MODE_REFSEL = _bv(5)
    MODE_NEGBUF = _bv(6)
    MODE_NOCHOP = _bv(7)

    CON_AIN1AINCOM = 0 << 4
    CON_AIN2AINCOM = 1 << 4
    CON_AIN3AINCOM = 2 << 4
    CON_AIN4AINCOM = 3 << 4
    CON_AIN5AINCOM = 4 << 4
    CON_AIN6AINCOM = 5 << 4
    CON_AIN7AINCOM = 6 << 4
    CON_AIN8AINCOM = 7 << 4
    CON_AIN1AIN2 = 8 << 4
    CON_AIN3AIN4 = 9 << 4
    CON_AIN5AIN6 = 10 << 4
    CON_AIN7AIN8 = 11 << 4
    CON_AIN2AIN2 = 12 << 4
    CON_AINCOMAINCOM = 13 << 4
    CON_REFINREFIN = 14 << 4
    CON_OPEN = 15 << 4
    CON_UNIPOLAR = _bv(3)

    CON_RANGE0 = 0  # +-20mV
    CON_RANGE1 = 1  # +-40mV
    CON_RANGE2 = 2  # +-80mV
    CON_RANGE3 = 3  # +-160mV
    CON_RANGE4 = 4  # +-320mV
    CON_RANGE5 = 5  # +-640mV
    CON_RANGE6 = 6  # +-1280mV
    CON_RANGE7 = 7  # +-2560mV
    gain = 1
    CHAN_NAMES = ['AIN1AINCOM', 'AIN2AINCOM', 'AIN3AINCOM', 'AIN4AINCOM', 'AIN5AINCOM',
                  'AIN6AINCOM', 'AIN7AINCOM', 'AIN8AINCOM']

    def __init__(self, I, calibs):
        self.cs = 'CS1'
        self.I = I
        self.calibs = calibs
        self.I.SPI.set_parameters(2, 1, 0, 1)
        self.writeRegister(self.FILTER, 20)
        self.writeRegister(self.MODE, self.MODE_SINGLE | self.MODE_CHCON | self.MODE_REFSEL)
        self.caldata = {}
        for a in calibs.keys():
            self.caldata[a] = np.poly1d(calibs[a])
        print('Loaded calibration', self.caldata)

    def start(self):
        self.I.SPI.start(self.cs)

    def stop(self):
        self.I.SPI.stop(self.cs)

    def send8(self, val):
        return self.I.SPI.send8(val)

    def send16(self, val):
        return self.I.SPI.send16(val)

    def write(self, regname, value):
        pass

    def readRegister(self, regname):
        self.start()
        val = self.send16(0x4000 | (regname << 8))
        self.stop()
        # print (regname,val)
        val &= 0x00FF
        return val

    def readData(self):
        self.start()
        val = self.send16(0x4000 | (self.ADCDATA << 8))
        val &= 0xFF
        val <<= 16
        val |= self.send16(0x0000)
        self.stop()
        return val

    def writeRegister(self, regname, value):
        self.start()
        val = self.send16(0x0000 | (regname << 8) | value)
        self.stop()
        return val

    def internalCalibration(self, chan=1):
        self.start()
        val = self.send16(0x0000 | (self.ADCCON << 8) | (chan << 4) | 7)  # range=7

        start_time = time.time()
        caldone = False
        val = self.send16(0x0000 | (self.MODE << 8) | 4)
        while caldone != 1:
            time.sleep(0.5)
            caldone = self.send16(0x4000 | (self.MODE << 8)) & 7
            print('waiting for zero scale calibration... %.2f S, %d' % (time.time() - start_time, caldone))

        print('\n')
        caldone = False
        val = self.send16(0x0000 | (self.MODE << 8) | 5)
        while caldone != 1:
            time.sleep(0.5)
            caldone = self.send16(0x4000 | (self.MODE << 8)) & 7
            print('waiting for full scale calibration... %.2f S %d' % (time.time() - start_time, caldone))

        print('\n')

        self.stop()

    def readCalibration(self):
        self.start()
        off = self.send16(0x4000 | (self.ADCOFFSET << 8))
        off &= 0xFF
        off <<= 16
        off |= self.send16(0x0000)

        gn = self.send16(0x4000 | (self.ADCGAIN << 8))
        gn &= 0xFF
        gn <<= 16
        gn |= self.send16(0x0000)
        self.stop()
        return off, gn

    def configADC(self, adccon):
        self.writeRegister(self.ADCCON, adccon)  # unipolar channels , range
        self.gain = 2 ** (7 - adccon & 3)

    def printstat(self):
        stat = self.readRegister(self.STATUS)
        P = ['PLL LOCKED', 'RES', 'RES', 'ADC ERROR', 'RES', 'CAL DONE', 'RES', 'READY']
        N = ['PLL ERROR', 'RES', 'RES', 'ADC OKAY', 'RES', 'CAL LOW', 'RES', 'NOT READY']
        s = ''
        for a in range(8):
            if stat & (1 << a):
                s += '\t' + P[a]
            else:
                s += '\t' + N[a]
        print(stat, s)

    def convert_unipolar(self, x):
        return (1.024 * self.VREF * x) / (self.gain * 2 ** 24)

    def convert_bipolar(self, x):
        return ((x / 2 ** 24) - 1) * (1.024 * self.VREF) / (self.gain)

    def __startRead__(self, chan):
        if chan not in self.CHAN_NAMES:
            print('invalid channel name. try AIN1AINCOM')
            return False
        chanid = self.CHAN_NAMES.index(chan)
        self.configADC(self.CON_RANGE7 | self.CON_UNIPOLAR | (chanid << 4))
        self.writeRegister(self.MODE, self.MODE_SINGLE | self.MODE_CHCON | self.MODE_REFSEL)
        return True

    def __fetchData__(self, chan):
        while True:
            stat = self.readRegister(self.STATUS)
            if stat & 0x80:
                data = float(self.readData())
                data = self.convert_unipolar(data)
                if int(chan[3]) > 4: data = (data - 3.3 / 2) * 4
                return self.caldata[chan](data)
            else:
                time.sleep(0.1)
                print('increase delay')
        return False

    def readVoltage(self, chan):
        if not self.__startRead__(chan):
            return False
        time.sleep(0.15)
        return self.__fetchData__(chan)

    def __fetchRawData__(self, chan):
        while True:
            stat = self.readRegister(self.STATUS)
            if stat & 0x80:
                data = float(self.readData())
                return self.convert_unipolar(data)
            else:
                time.sleep(0.01)
                print('increase delay')
        return False

    def readRawVoltage(self, chan):
        if not self.__startRead__(chan):
            return False
        time.sleep(0.15)
        return self.__fetchRawData__(chan)


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    calibs = {
        'AIN6AINCOM': [6.993123e-07, -1.563294e-06, 9.994211e-01, -4.596018e-03],
        'AIN7AINCOM': [3.911521e-07, -1.706405e-06, 1.002294e+00, -1.286302e-02],
        'AIN3AINCOM': [-3.455831e-06, 2.861689e-05, 1.000195e+00, 3.802349e-04],
        'AIN1AINCOM': [8.220199e-05, -4.587100e-04, 1.001015e+00, -1.684517e-04],
        'AIN5AINCOM': [-1.250787e-07, -9.203838e-07, 1.000299e+00, -1.262684e-03],
        'AIN2AINCOM': [5.459186e-06, -1.749624e-05, 1.000268e+00, 1.907896e-04],
        'AIN9AINCOM': [7.652808e+00, 1.479229e+00, 2.832601e-01, 4.495232e-02],
        'AIN8AINCOM': [8.290843e-07, -7.129532e-07, 9.993159e-01, 3.307947e-03],
        'AIN4AINCOM': [4.135213e-06, -1.973478e-05, 1.000277e+00, 2.115374e-04], }
    A = AD7718(I, calibs)
    for a in range(10):
        print(A.readRawVoltage('AIN1AINCOM'))
        time.sleep(0.3)



import inspect

from PSL.SENSORS import HMC5883L
from PSL.SENSORS import MPU6050
from PSL.SENSORS import MLX90614
from PSL.SENSORS import BMP180
from PSL.SENSORS import TSL2561
from PSL.SENSORS import SHT21
from PSL.SENSORS import BH1750
from PSL.SENSORS import SSD1306

supported = {
    0x68: MPU6050,  # 3-axis gyro,3-axis accel,temperature
    0x1E: HMC5883L,  # 3-axis magnetometer
    0x5A: MLX90614,  # Passive IR temperature sensor
    0x77: BMP180,  # Pressure, Temperature, altitude
    0x39: TSL2561,  # Luminosity
    0x40: SHT21,  # Temperature, Humidity
    0x23: BH1750,  # Luminosity
    # 0x3C:SSD1306,    #OLED display
}

# auto generated map of names to classes
nameMap = {}
for a in supported:
    nameMap[supported[a].__name__.split('.')[-1]] = (supported[a])



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pslab",
    version="2.2.0",
    description="Pocket Science Lab by FOSSASIA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="FOSSASIA PSLab Developers",
    author_email="pslab-fossasia@googlegroups.com",
    url="https://pslab.io/",
    install_requires=["numpy>=1.16.3", "pyserial>=3.4", "scipy>=1.3.0"],
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=("tests",)),
    package_data={
        "": [
            "*.css",
            "*.png",
            "*.gif",
            "*.html",
            "*.css",
            "*.js",
            "*.png",
            "*.jpg",
            "*.jpeg",
            "*.htm",
            "99-pslab.rules",
        ]
    },
    entry_points = {
        'console_scripts': ['pslab=pslab.cli:cmdline'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
)



"""Control voltage and current with the PSLab's PV1, PV2, PV3, and PCS pins.

Examples
--------
>>> from pslab import PowerSupply
>>> ps = PowerSupply()
>>> ps.pv1 = 4.5
>>> ps.pv1
4.499389499389499

>>> ps.pcs = 2e-3
>>> ps.pcs
0.00200014652014652
"""
import numpy as np

from pslab.bus.i2c import I2CSlave
from pslab.serial_handler import SerialHandler


class PowerSupply:
    """Control the PSLab's programmable voltage and current sources.

    An instance of PowerSupply controls three programmable voltage sources on
    pins PV1, PV2, and PV3, as well as a programmable current source on pin
    PCS. The voltage/current on each source can be set via the voltage/current
    properties of each source.

    Parameters
    ----------
    device : :class:`SerialHandler`
        Serial connection with which to communicate with the device. A new
        instance will be created automatically if not specified.
    """

    _ADDRESS = 0x60

    def __init__(self, device: SerialHandler = None):
        self._device = device if device is not None else SerialHandler()
        self._mcp4728 = I2CSlave(self._ADDRESS, self._device)
        self._pv1 = VoltageSource(self._mcp4728, "PV1")
        self._pv2 = VoltageSource(self._mcp4728, "PV2")
        self._pv3 = VoltageSource(self._mcp4728, "PV3")
        self._pcs = CurrentSource(self._mcp4728)

    @property
    def pv1(self):
        """float: Voltage on PV1; range [-5, 5] V."""
        return self._pv1.voltage

    @pv1.setter
    def pv1(self, value: float):
        self._pv1.voltage = value

    @property
    def pv2(self):
        """float: Voltage on PV2; range [-3.3, 3.3] V."""
        return self._pv2.voltage

    @pv2.setter
    def pv2(self, value: float):
        self._pv2.voltage = value

    @property
    def pv3(self):
        """float: Voltage on PV3; range [0, 3.3] V."""
        return self._pv3.voltage

    @pv3.setter
    def pv3(self, value: float):
        self._pv3.voltage = value

    @property
    def pcs(self):
        """float: Current on PCS; range [0, 3.3e-3] A.

        Notes
        -----
        The maximum available current that can be output by the current source
        is dependent on load resistance:

            I_max = 3.3 V / (1 kΩ + R_load)

        For example, the maximum current that can be driven across a 100 Ω load
        is 3.3 V / 1.1 kΩ = 3 mA. If the load is 10 kΩ, the maximum current is
        only 3.3 V / 11 kΩ = 300 µA.

        Be careful to not set a current higher than available for a given load.
        If a current greater than the maximum for a certain load is requested,
        the actual current will instead be much smaller. For example, if a
        current of 3 mA is requested when connected to a 1 kΩ load, the actual
        current will be only a few hundred µA instead of the maximum available
        1.65 mA.
        """
        return self._pcs.current

    @pcs.setter
    def pcs(self, value: float):
        self._pcs.current = value

    @property
    def _registers(self):
        """Return the contents of the MCP4728's input registers and EEPROM."""
        return self._mcp4728.read(24)


class Source:
    """Base class for voltage/current/power sources."""

    _RANGE = {
        "PV1": (-5, 5),
        "PV2": (-3.3, 3.3),
        "PV3": (0, 3.3),
        "PCS": (3.3e-3, 0),
    }
    _CHANNEL_NUMBER = {
        "PV1": 3,
        "PV2": 2,
        "PV3": 1,
        "PCS": 0,
    }
    _RESOLUTION = 2 ** 12 - 1
    _MULTI_WRITE = 0b01000000

    def __init__(self, mcp4728: I2CSlave, name: str):
        self._mcp4728 = mcp4728
        self.name = name
        self.channel_number = self._CHANNEL_NUMBER[self.name]
        slope = self._RANGE[self.name][1] - self._RANGE[self.name][0]
        intercept = self._RANGE[self.name][0]
        self._unscale = np.poly1d(
            [self._RESOLUTION / slope, -self._RESOLUTION * intercept / slope]
        )
        self._scale = np.poly1d([slope / self._RESOLUTION, intercept])

    def unscale(self, voltage: float) -> int:
        """Return integer representation of a voltage.

        Parameters
        ----------
        voltage : float
            Voltage in Volt.

        Returns
        -------
        raw : int
            Integer represention of the voltage.
        """
        return int(round(self._unscale(voltage)))

    def scale(self, raw: int) -> float:
        """Convert an integer value to a voltage value.

        Parameters
        ----------
        raw : int
            Integer representation of a voltage value.

        Returns
        -------
        voltage : float
            Voltage in Volt.
        """
        return self._scale(raw)

    def _multi_write(self, raw: int):
        channel_select = self.channel_number << 1
        command_byte = self._MULTI_WRITE | channel_select
        data_byte1 = (raw >> 8) & 0x0F
        data_byte2 = raw & 0xFF
        self._mcp4728.write([data_byte1, data_byte2], register_address=command_byte)


class VoltageSource(Source):
    """Helper class for interfacing with PV1, PV2, and PV3."""

    def __init__(self, mcp4728: I2CSlave, name: str):
        self._voltage = 0
        super().__init__(mcp4728, name)

    @property
    def voltage(self):
        """float: Most recent voltage set on PVx in Volt.

        The voltage on PVx can be set by writing to this attribute.
        """
        return self._voltage

    @voltage.setter
    def voltage(self, value: float):
        raw = self.unscale(value)
        raw = int(np.clip(raw, 0, self._RESOLUTION))
        self._multi_write(raw)
        self._voltage = self.scale(raw)


class CurrentSource(Source):
    """Helper class for interfacing with PCS."""

    def __init__(self, mcp4728: I2CSlave):
        self._current = 0
        super().__init__(mcp4728, "PCS")

    @property
    def current(self):
        """float: Most recent current value set on PCS in Ampere.

        The current on PCS can be set by writing to this attribute.
        """
        return self._current

    @current.setter
    def current(self, value: float):
        # Output current is a function of the voltage set on the MCP4728's V+
        # pin (assuming negligible load resistance):
        #    I(V) = 3.3e-3 - V / 1000
        # I.e. the lower the voltage the higher the current. However, the
        # function is discontinuous in V = 0:
        #    I(0) = 0
        if value == 0:
            self._multi_write(0)
            self._current = 0
        else:
            raw = self.unscale(value)
            raw = int(np.clip(raw, 0, self._RESOLUTION))
            self._multi_write(raw)
            self._current = self.scale(raw)



"""Control the PSLab's waveform generators.

Two types of waveform generator are available: WaveformGenerator and
PWMGenerator. WaveformGenerator can output arbitrary waveforms on pins SI1 and
SI2. PWMGenerator can output square waveforms on pins SQ1, SQ2, SQ3, and SQ4.
"""

import logging
from typing import Callable, List, Tuple, Union

import numpy as np

import pslab.protocol as CP
from pslab.instrument.analog import AnalogOutput
from pslab.instrument.digital import DigitalOutput, DIGITAL_OUTPUTS
from pslab.serial_handler import SerialHandler

logger = logging.getLogger(__name__)

_PRESCALERS = [1, 8, 64, 256]


def _listify(channel, maxlen, *args):
    if not isinstance(channel, list):
        channel = [channel]
    elif len(channel) > maxlen:
        raise ValueError("Too many channels.")

    ret = [channel]

    for arg in args:
        if isinstance(arg, list):
            if len(arg) == len(channel):
                ret.append(arg)
            else:
                raise ValueError("Dimension mismatch.")
        else:
            ret.append(len(channel) * [arg])

    return ret


def _get_wavelength(frequency: float, table_size: int = 1) -> Tuple[int, int]:
    """Get the wavelength of a PWM signal in clock cycles, and the clock prescaler.

    For an analog signal, the wavelength of the underlying PWM signal is equal
    to the time gap between two points in the analog waveform.

    Parameters
    ----------
    frequency : float
        Frequency of the signal in Hz.
    table_size : int, optional
        Number of points in the analog signal. The default value is 1, which
        implies a digital signal.

    Returns
    -------
    wavelength : int
        Signal wavelength in clock cycles.
    prescaler : int
        Factor by which the clock rate is reduced, e.g. a prescaler of 64 means
        that the clock rate is 1 Mhz instead of the original 64 MHz.
    """
    for prescaler in _PRESCALERS:
        timegap = int(round(CP.CLOCK_RATE / frequency / prescaler / table_size))
        if 0 < timegap < 2 ** 16:
            return timegap, prescaler

    e = (
        "Prescaler out of range. This should not happen."
        + " "
        + "Please report this bug, including steps to trigger it, to"
        + " "
        + "https://github.com/fossasia/pslab-python/issues."
    )
    raise ValueError(e)


class WaveformGenerator:
    """Generate analog waveforms on SI1 or SI2.

    Parameters
    ----------
    device : :class:`SerialHandler`, optional
        Serial connection with which to communicate with the device. A new
        instance is created automatically if not specified.

    Examples
    --------
    Output the default function (3.3*sin(t)) on SI2 with frequency 2.5 kHz:

    >>> from pslab import WaveformGenerator
    >>> wavegen = WaveformGenerator()
    >>> wavegen.generate("SI2", 2500)
    [2500]

    Output phase shifted sine waves on SI1 and SI2:

    >>> wavegen.generate(["SI1", "SI2"], 1000, 90)
    [1000, 1000]

    Reduce the amplitude on SI1:

    >>> import numpy as np
    >>> wavegen.load_function("SI1", lambda x: 1.5*np.sin(x), [0, 2*np.pi])

    Output two superimposed sine waves on SI2:

    >>> wavegen.load_function("SI2", lambda x: 2*np.sin(x) + np.sin(5*x), [0, 2*np.pi])
    """

    _HIGHRES_TABLE_SIZE = 512
    _LOWRES_TABLE_SIZE = 32
    _LOW_FREQUENCY_WARNING = 20
    _LOW_FREQUENCY_LIMIT = 0.1
    _HIGH_FREQUENCY_WARNING = 5e3
    _HIGHRES_FREQUENCY_LIMIT = 1100

    def __init__(self, device: SerialHandler = None):
        self._channels = {n: AnalogOutput(n) for n in ("SI1", "SI2")}
        self._device = device if device is not None else SerialHandler()

    def generate(
        self,
        channels: Union[str, List[str]],
        frequency: Union[float, List[float]],
        phase: float = 0,
    ) -> List[float]:
        """Generate analog waveforms on SI1 or SI2.

        The default waveform is a sine wave with amplitude 3.3 V. Other
        waveforms can be set using :meth:`load_function` or :meth:`load_table`.

        Parameters
        ----------
        channels : {1, 2} or {'SI1', 'SI2', ['SI1', 'SI2']}
            Pin(s) on which to generate a waveform.
        frequency : float or list of floats
            Frequency in Hz. Can be a list containing two different values when
            'channel' is ['SI1', 'SI2']. Must be greater than 0.1 Hz. For
            frequencies below 1 Hz the signal is noticably attenuated by AC
            coupling.
        phase : float, optional
            Phase between waveforms when generating waveforms on both SI1 and
            SI2 in degrees. The default is 0.

        Returns
        -------
        frequency : List[float]
            The actual frequency may differ from the requested value due to
            the device-interal integer representation. The actual frequency is
            therefore returned as a list. The length of the list is equal to
            the number of channels used to generate waveforms.
        """
        if isinstance(channels, int):
            channels = ["SI1", "SI2"][:channels]

        channels, frequency = _listify(channels, 2, frequency)
        table_size = len(channels) * [None]
        timegap = len(channels) * [None]
        prescaler = len(channels) * [None]

        for i, (chan, freq) in enumerate(zip(channels, frequency)):
            if freq < self._LOW_FREQUENCY_WARNING:
                w = (
                    f"Frequencies under {self._LOW_FREQUENCY_WARNING} Hz have"
                    + " "
                    + "reduced amplitude due to AC coupling restrictions."
                )
                logger.warning(w)
            elif freq > self._HIGH_FREQUENCY_WARNING:
                w = (
                    f"Frequencies above {self._HIGH_FREQUENCY_WARNING} Hz have"
                    + " "
                    + "reduced amplitude."
                )
                logger.warning(w)

            table_size[i] = self._get_table_size(freq)
            timegap[i], prescaler[i] = _get_wavelength(freq, table_size[i])
            frequency[i] = CP.CLOCK_RATE / timegap[i] / prescaler[i] / table_size[i]
            self._channels[chan].frequency = frequency[i]

        if len(channels) == 1:
            self._output_one(channels, table_size, prescaler, timegap)
        else:
            self._output_two(table_size, phase, prescaler, timegap)

        return frequency

    def _output_one(self, channel: str, table_size: int, prescaler: int, timegap: int):
        self._device.send_byte(CP.WAVEGEN)
        secondary_cmd = CP.SET_SINE1 if channel[0] == "SI1" else CP.SET_SINE2
        self._device.send_byte(secondary_cmd)
        # Use larger table for low frequencies.
        highres = table_size[0] == self._HIGHRES_TABLE_SIZE
        self._device.send_byte(highres | (_PRESCALERS.index(prescaler[0]) << 1))
        self._device.send_int(timegap[0] - 1)
        self._device.get_ack()

    def _output_two(self, table_size: int, phase: float, prescaler: int, timegap: int):
        self._device.send_byte(CP.WAVEGEN)
        phase_coarse = int(table_size[1] * phase / 360)
        phase_fine = int(
            timegap[1]
            * (phase - phase_coarse * 360 / table_size[1])
            / (360 / table_size[1])
        )
        self._device.send_byte(CP.SET_BOTH_WG)
        self._device.send_int(timegap[0] - 1)
        self._device.send_int(timegap[1] - 1)
        self._device.send_int(phase_coarse)  # Table position for phase adjust.
        self._device.send_int(phase_fine)  # Timer delay / fine phase adjust.
        highres = [t == self._HIGHRES_TABLE_SIZE for t in table_size]
        self._device.send_byte(
            (_PRESCALERS.index(prescaler[1]) << 4)
            | (_PRESCALERS.index(prescaler[0]) << 2)
            | (highres[1] << 1)
            | (highres[0])
        )
        self._device.get_ack()

    def _get_table_size(self, frequency: float) -> int:
        if frequency < self._LOW_FREQUENCY_LIMIT:
            e = f"Frequency must be greater than {self._LOW_FREQUENCY_LIMIT} Hz."
            raise ValueError(e)
        elif frequency < self._HIGHRES_FREQUENCY_LIMIT:
            table_size = self._HIGHRES_TABLE_SIZE
        else:
            table_size = self._LOWRES_TABLE_SIZE

        return table_size

    def load_function(
        self, channel: str, function: Union[str, Callable], span: List[float] = None
    ):
        """Load a custom waveform function.

        Parameters
        ----------
        channel : {'SI1', 'SI2'}
            The output pin on which to generate the waveform.
        function : Union[str, Callable]
            A callable function which takes a numpy.ndarray of x values as
            input and returns a corresponding numpy.ndarray of y values. The
            y-values should be voltages in V, and should lie between -3.3 V
            and 3.3 V. Values outside this range will be clipped.

            Alternatively, 'function' can be a string literal 'sine' or 'tria',
            for a sine wave or triangle wave with amplitude 3.3 V.
        span : List[float], optional
            The minimum and maximum x values between which to evaluate
            'function'. Should typically correspond to one period. Omit if
            'function' is 'sine' or 'tria'.
        """
        if function == "sine":

            def sine(x):
                return AnalogOutput.RANGE[1] * np.sin(x)

            function = sine
            span = [0, 2 * np.pi]
            self._channels[channel].wavetype = "sine"
        elif function == "tria":

            def tria(x):
                return AnalogOutput.RANGE[1] * (abs(x % 4 - 2) - 1)

            function = tria
            span = [-1, 3]
            self._channels[channel].wavetype = "tria"
        else:
            self._channels[channel].wavetype = "custom"

        x = np.arange(span[0], span[1], (span[1] - span[0]) / 512)
        y = function(x)
        self._load_table(
            channel=channel, points=y, mode=self._channels[channel].wavetype
        )

    def load_table(self, channel: str, points: np.ndarray):
        """Load a custom waveform as a table.

        Parameters
        ----------
        channel : {'SI1', 'SI2'}
            The output pin on which to generate the waveform.
        points : np.ndarray
            Array of voltage values which make up the waveform. Array length
            must be 512. Values outside the range -3.3 V to 3.3 V will be
            clipped.
        """
        self._load_table(channel, points, "custom")

    def _load_table(self, channel, points, mode="custom"):
        self._channels[channel].wavetype = mode
        self._channels[channel].waveform_table = points
        logger.info(f"Reloaded waveform table for {channel}: {mode}.")
        self._device.send_byte(CP.WAVEGEN)

        if self._channels[channel].name == "SI1":
            self._device.send_byte(CP.LOAD_WAVEFORM1)
        else:
            self._device.send_byte(CP.LOAD_WAVEFORM2)

        for val in self._channels[channel].waveform_table:
            self._device.send_int(val)
        for val in self._channels[channel].lowres_waveform_table:
            self._device.send_byte(val)

        self._device.get_ack()


class PWMGenerator:
    """Generate PWM signals on SQ1, SQ2, SQ3, and SQ4.

    Parameters
    ----------
    device : :class:`SerialHandler`
        Serial connection with which to communicate with the device. A new
        instance will be created automatically if not specified.

    Examples
    --------
    Output 40 kHz PWM signals on SQ1 and SQ3 phase shifted by 50%. Set the duty
    cycles to 75% and 33%, respectivelly:

    >>> from pslab import PWMGenerator
    >>> pwmgen = PWMGenerator()
    >>> pwmgen.generate(["SQ1", "SQ2"], 4e4, [0.75, 0.33], 0.5)

    Output a 32 MHz PWM signal on SQ4 with a duty cycle of 50%:

    >>> pwmgen.map_reference_clock(["SQ4"], 2)

    Set SQ2 high:

    >>> pwmgen.set_states(sq2=True)
    """

    _HIGH_FREQUENCY_LIMIT = 1e7

    def __init__(self, device: SerialHandler = None):
        self._device = device if device is not None else SerialHandler()
        self._channels = {n: DigitalOutput(n) for n in DIGITAL_OUTPUTS}
        self._frequency = 0
        self._reference_prescaler = 0

    @property
    def frequency(self) -> float:
        """Get the common frequency for all digital outputs in Hz."""
        return self._frequency

    def generate(
        self,
        channels: Union[str, List[str]],
        frequency: float,
        duty_cycles: Union[float, List[float]],
        phases: Union[float, List[float]] = 0,
    ):
        """Generate PWM signals on SQ1, SQ2, SQ3, and SQ4.

        Parameters
        ----------
        channels : {1, 2, 3, 4} or {'SQ1', 'SQ2', 'SQ3', 'SQ4'} or list of the same
            Pin name or list of pin names on which to generate PWM signals.
            Pins which are not included in the argument will not be affected.
        frequency : float
            Frequency in Hz. Shared by all outputs.
        duty_cycles : float or list of floats
            Duty cycle between 0 and 1 as either a single value or a list of
            values.

            If 'duty_cycles' is a single value, it is applied to all channels
            given in 'channels'.

            If 'duty_cycles' is a list, the values in the list will be applied
            to the corresponding channel in the 'channels' list. The lists must
            have the same length.
        phases : float or list of floats
            Phase between 0 and 1 as either a single value or a list of values.

            If 'phases' is a single value, it will be the phase between each
            subsequent channel in 'channels'. For example,

            >>> generate(['SQ1', 'SQ2', 'SQ3', 'SQ4'], 1000, 0.5, 0.1)

            SQ2 will be shifted relative to SQ1 by 10%, SQ3 will be shifted
            relative to SQ2 by 10% (i.e. 20% relative to SQ1), and SQ4 will be
            shifted relative to SQ3 by 10% (i.e. 30% relative to SQ1).

            If 'phases' is a list, the values in the list will be applied to
            the corresponding channel in the 'channels' list. The lists must
            have the same length.
        """
        if isinstance(channels, int):
            channels = ["SQ1", "SQ2", "SQ3", "SQ4"][:channels]

        if frequency > self._HIGH_FREQUENCY_LIMIT:
            e = (
                "Frequency is greater than 10 MHz."
                + " "
                + "Please use map_reference_clock for 16 & 32 MHz outputs."
            )
            raise ValueError(e)
        elif frequency <= 0:
            raise ValueError("Frequency must be positive.")
        else:
            self._frequency = frequency
            channels, duty_cycles = _listify(channels, 4, duty_cycles)

            if not isinstance(phases, list):
                phases = [i * phases for i in range(len(channels))]

            for channel, duty_cycle, phase in zip(channels, duty_cycles, phases):
                self._channels[channel].duty_cycle = duty_cycle
                self._channels[channel].phase = phase
                self._channels[channel].remapped = False

            # Turn on all channels, minimum duty cycle of 1 wavelength.
            self._generate(
                [self._channels[c].duty_cycle for c in DIGITAL_OUTPUTS],
                [self._channels[c].phase for c in DIGITAL_OUTPUTS],
            )
            # Reset channels which should be LOW or HIGH to correct duty cycle.
            self.set_state(**{k.lower(): v.state for k, v in self._channels.items()})

            # Remap channels which should be mapped to the referene clock.
            remapped = [c.name for c in self._channels.values() if c.remapped]
            self.map_reference_clock(remapped, self._reference_prescaler)

    def _generate(
        self,
        duty_cycles: List[float],
        phases: List[float],
    ):
        """Generate PWM signals on all four digital output pins.

        Paramaters
        ----------
        duty_cycles : list of floats
            List of length four containing duty cycles for each output as a
            float in the open interval (0, 1). Note that it is not possible
            to set the duty cycle to exactly 0 or 1; use :meth:`set_state`
            instead.
        phases : list of floats
            List of length four containing phases for each output as a float in
            the half open interval [0, 1). A phase of 1 is identical to a phase
            of 0.
        """
        wavelength, prescaler = _get_wavelength(self._frequency)
        self._frequency = CP.CLOCK_RATE / wavelength / prescaler
        continuous = 1 << 5

        for i, (duty_cycle, phase) in enumerate(zip(duty_cycles, phases)):
            duty_cycles[i] = int((duty_cycle + phase) % 1 * wavelength)
            duty_cycles[i] = max(1, duty_cycles[i] - 1)  # Zero index.
            phases[i] = int(phase % 1 * wavelength)
            phases[i] = max(0, phases[i] - 1)  # Zero index.

        self._device.send_byte(CP.WAVEGEN)
        self._device.send_byte(CP.SQR4)
        self._device.send_int(wavelength - 1)  # Zero index.
        self._device.send_int(duty_cycles[0])
        self._device.send_int(phases[1])
        self._device.send_int(duty_cycles[1])
        self._device.send_int(phases[2])
        self._device.send_int(duty_cycles[2])
        self._device.send_int(phases[3])
        self._device.send_int(duty_cycles[3])
        self._device.send_byte(_PRESCALERS.index(prescaler) | continuous)
        self._device.get_ack()

    def set_state(
        self,
        sq1: Union[bool, str, None] = None,
        sq2: Union[bool, str, None] = None,
        sq3: Union[bool, str, None] = None,
        sq4: Union[bool, str, None] = None,
    ):
        """Set the digital outputs HIGH or LOW.

        Parameters
        ----------
        sq1 : {True, False, None, 'HIGH', 'LOW', 'PWM'}, optional
            Set the state of SQ1. True or "HIGH" sets it HIGH, False or "LOW"
            sets it low, and None or "PWM" leaves it in its current state. The
            default value is None.
        sq2 : {True, False, None, 'HIGH', 'LOW', 'PWM'}, optional
            See 'sq1'.
        sq3 : {True, False, None, 'HIGH', 'LOW', 'PWM'}, optional
            See 'sq1'.
        sq4 : {True, False, None, 'HIGH', 'LOW', 'PWM'}, optional
            See 'sq1'.
        """
        states = 0

        for i, sq in enumerate([sq1, sq2, sq3, sq4]):
            if sq in (True, False, "HIGH", "LOW"):
                sq = 1 if sq in (True, "HIGH") else 0
                self._channels[DIGITAL_OUTPUTS[i]].duty_cycle = sq
                states |= self._channels[DIGITAL_OUTPUTS[i]].state_mask | (sq << i)

        self._device.send_byte(CP.DOUT)
        self._device.send_byte(CP.SET_STATE)
        self._device.send_byte(states)
        self._device.get_ack()

    def map_reference_clock(self, channels: List[str], prescaler: int):
        """Map the internal oscillator output to a digital output.

        The duty cycle of the output is locked to 50%.

        Parameters
        ----------
        channels : {'SQ1', 'SQ2', 'SQ3', 'SQ4'} or list of the same
            Digital output pin(s) to which to map the internal oscillator.
        prescaler : int
            Prescaler value in interval [0, 15]. The output frequency is
            128 / (1 << prescaler) MHz.
        """
        (channels,) = _listify(channels, 4)
        self._device.send_byte(CP.WAVEGEN)
        self._device.send_byte(CP.MAP_REFERENCE)
        self._reference_prescaler = prescaler
        maps = 0

        for channel in channels:
            self._channels[channel].duty_cycle = 0.5
            self._channels[channel].phase = 0
            self._channels[channel].remapped = True
            maps |= self._channels[channel].reference_clock_map

        self._device.send_byte(maps)
        self._device.send_byte(prescaler)
        self._device.get_ack()



"""Objects related to the PSLab's digital input channels."""

import numpy as np

DIGITAL_INPUTS = ("LA1", "LA2", "LA3", "LA4", "RES", "EXT", "FRQ")

DIGITAL_OUTPUTS = ("SQ1", "SQ2", "SQ3", "SQ4")

MODES = {
    "sixteen rising": 5,
    "four rising": 4,
    "rising": 3,
    "falling": 2,
    "any": 1,
    "disabled": 0,
}


class DigitalInput:
    """Model of the PSLab's digital inputs.

    Parameters
    ----------
    name : {"LA1", "LA2", "LA3", "LA4", "RES", "FRQ"}
        Name of the digital channel to model.

    Attributes
    ----------
    name : str
        One of {"LA1", "LA2", "LA3", "LA4", "RES", "FRQ"}.
    number : int
        Number used to refer to this channel in the firmware.
    datatype : str
        Either "int" or "long", depending on if a 16 or 32-bit counter is used
        to capture timestamps for this channel.
    events_in_buffer : int
        Number of logic events detected on this channel, the timestamps of
        which are currently being held in the device's ADC buffer.
    buffer_idx : Union[int, None]
        Location in the device's ADC buffer where the events are stored. None
        if no events captured by this channel are currently held in the buffer.
    """

    def __init__(self, name: str):
        self.name = name
        self.number = DIGITAL_INPUTS.index(self.name)
        self.datatype = "long"
        self.events_in_buffer = 0
        self.buffer_idx = None
        self._logic_mode = MODES["any"]

    @property
    def logic_mode(self) -> str:
        """str: Type of logic event which should be captured on this channel.

        The options are:
                any:            Capture every edge.
                rising:         Capture every rising edge.
                falling:        Capture every falling edge.
                four rising:    Capture every fourth rising edge.
                sixteen rising: Capture every fourth rising edge.
        """
        return {v: k for k, v in MODES.items()}[self._logic_mode]

    def _get_xy(self, initial_state: bool, timestamps: np.ndarray):
        x = np.repeat(timestamps, 3)
        x = np.insert(x, 0, 0)
        x[0] = 0
        y = np.array(len(x) * [False])

        if self.logic_mode == "any":
            y[0] = initial_state
            for i in range(1, len(x), 3):
                y[i] = y[i - 1]  # Value before this timestamp.
                y[i + 1] = not y[i]  # Value at this timestamp.
                y[i + 2] = y[i + 1]  # Value leaving this timetamp.
        elif self.logic_mode == "falling":
            y[0] = True
            for i in range(1, len(x), 3):
                y[i] = True  # Value before this timestamp.
                y[i + 1] = False  # Value at this timestamp.
                y[i + 2] = True  # Value leaving this timetamp.
        else:
            y[0] = False
            for i in range(1, len(x), 3):
                y[i] = False  # Value before this timestamp.
                y[i + 1] = True  # Value at this timestamp.
                y[i + 2] = False  # Value leaving this timetamp.

        return x, y


class DigitalOutput:
    """Model of the PSLab's digital outputs.

    Parameters
    ----------
    name : {'SQ1', 'SQ2', 'SQ3', 'SQ4'}
        Name of the digital output pin represented by the instance.
    """

    def __init__(self, name: str):
        self._name = name
        self._state = "LOW"
        self._duty_cycle = 0
        self.phase = 0
        self.remapped = False

    @property
    def name(self) -> str:
        """str: Name of this pin."""
        return self._name

    @name.setter
    def name(self, value):
        if value in DIGITAL_OUTPUTS:
            self._name = value
        else:
            e = f"Invalid digital output {value}. Choose one of {DIGITAL_OUTPUTS}."
            raise ValueError(e)

    @property
    def state(self) -> str:
        """str: State of the digital output. Can be 'HIGH', 'LOW', or 'PWM'."""
        return self._state

    @property
    def duty_cycle(self) -> float:
        """float: Duty cycle of the PWM signal on this pin."""
        return self._duty_cycle

    @duty_cycle.setter
    def duty_cycle(self, value: float):
        if value == 0:
            self._state = "LOW"
        elif value < 1:
            self._state = "PWM"
        elif value == 1:
            self._state = "HIGH"
        else:
            raise ValueError("Duty cycle must be in range [0, 1].")

        self._duty_cycle = value

    @property
    def state_mask(self) -> int:
        """int: State mask for this pin.

        The state mask is used in the DOUT->SET_STATE command to set the
        digital output pins HIGH or LOW. For example:

            0x10 | 1 << 0 | 0x40 | 0 << 2 | 0x80 | 1 << 3

        would set SQ1 and SQ4 HIGH, SQ3 LOW, and leave SQ2 unchanged.
        """
        if self.name == "SQ1":
            return 0x10
        elif self.name == "SQ2":
            return 0x20
        elif self.name == "SQ3":
            return 0x40
        elif self.name == "SQ4":
            return 0x80

    @property
    def reference_clock_map(self) -> int:
        """int: Reference clock map value for this pin.

        The reference clock map is used in the WAVEGEN->MAP_REFERENCE command
        to map a digital output pin directly to the interal oscillator. This
        can be used to achieve very high frequencies, with the caveat that
        the only frequencies available are quotients of 128 MHz and powers of
        2 up to 15. For example, sending (2 | 4) followed by 3 outputs
        128 / (1 << 3) = 16 MHz on SQ2 and SQ3.
        """
        if self.name == "SQ1":
            return 1
        elif self.name == "SQ2":
            return 2
        elif self.name == "SQ3":
            return 4
        elif self.name == "SQ4":
            return 8
